from abc import ABC, abstractmethod
from typing import Any, Generator, Optional

import simpy

from logis.biz.sim.const import GoHomeStrategyFrequency
from logis.biz.sim.graph import IPathGraph
from logis.data_type.base import Predicate

from .base import *


class AgentIdleStrategy(ABC):
    """
    智能体空闲策略
    """

    def __init__(self, agent: Optional[AgentClass] = None) -> None:
        """
        初始化智能体空闲策略
        Args:
            agent: 智能体,是否要传取决于此策略是否仅针对此智能体
        """
        super().__init__()
        self.agent = agent

    def is_idle(self, agent: Optional[AgentClass] = None, **kwargs) -> bool:
        """
        判断智能体是否空闲
        """
        raise NotImplementedError("is_idle方法尚未实现")

    @abstractmethod
    def on_idle(
        self,
        agent: Optional[AgentClass] = None,
        is_idle: Optional[Predicate[AgentClass]] = None,
        **kwargs,
    ) -> Generator[simpy.Event, Any, None]:
        """
        智能体做完任务之后的回调逻辑
        Args:
            agent: 如果不传则使用初始化时传的智能体，否则使用传参的智能体
            is_idle: 判断智能体是否空闲的方法，如果不传则使用is_idle方法，否则使用传参的方法判断
        """


class GoHomeStrategy(AgentIdleStrategy):
    """
    回到归属地策略
    """

    def __init__(
        self,
        frequency: GoHomeStrategyFrequency,
        agent: Optional[AgentClass] = None,
        env: Optional[simpy.Environment] = None,
        **kwargs,
    ) -> None:
        super().__init__(agent=agent, **kwargs)
        self.frequency: GoHomeStrategyFrequency = frequency
        self.env: simpy.Environment = env

    def on_idle(
        self,
        agent: Optional[AgentClass] = None,
        is_idle: Optional[Predicate[AgentClass]] = None,
        **kwargs,
    ):
        """
        默认空闲策略，智能体在空闲时，等待下一个任务
        """
        agent = agent or self.agent
        is_idle = is_idle or self.is_idle

        if (
            self.frequency == "如果无其他任务" and is_idle(agent, **kwargs)
        ) or self.frequency == "每次":
            yield from agent.move(target=agent.origin_location, **kwargs)
        else:
            raise NotImplementedError(f"频率{self.frequency}尚未支持")



from numbers import Number
from typing import Optional

import simpy

from logis.data_type import QuantifiedValue, Unit


class QuantifiedContainer:
    """
    初衷是统一simpy的Resource、Container、Store等类型并增加单位的概念，但目前还未完全实现

    引入单位的概念，能够简化对simpy的操作，可实现单位换算、检验等功能

    """

    __container__: Optional[simpy.Container] = None

    def __init__(
        self, capacity: QuantifiedValue, env: simpy.Environment, *args, **kwargs
    ):
        self.__quantified_value__ = capacity
        self.__container__ = simpy.Container(env, capacity.quantity)

    @property
    def unit(self) -> Unit:
        return self.__quantified_value__.unit

    @property
    def capacity(self):
        return self.__quantified_value__.quantity

    @property
    def level(self):
        return self.__container__.level

    @property
    def free_capacity(self) -> Number:
        """
        空闲空间
        """
        return self.capacity - self.level

    def get(self, v: QuantifiedValue):
        assert v.unit == self.unit, "unit must be the same"
        return self.__container__.get(v.quantity)

    def put(self, v: QuantifiedValue):
        assert v.unit == self.unit, "unit must be the same"
        return self.__container__.put(v.quantity)



import json
import time
from typing import Any, Dict, Generic, Optional, TypeVar

import pytest
from pydantic import BaseModel

from logis.data_type import Number
from logis.data_type.unitable import (
    Capacity,
    Length,
    NumberUnit,
    QuantifiedValue,
    Speed,
    SpeedVector,
    merge_quantified_value,
)


def test_number():
    assert isinstance(10, Number)
    assert isinstance(1.1, Number)


def test_capacity():
    x = Capacity.parse_tuple([10, "箱"])
    assert isinstance(x, Capacity)
    x = Capacity.parse_tuple([10, "箱"])
    assert isinstance(x, Capacity)
    assert isinstance(x, NumberUnit)
    assert isinstance(x, QuantifiedValue)


def test_speed():
    x = Speed.parse_str("1|m/s")
    assert isinstance(x, Speed)
    assert x.quantity == 1
    assert x.unit == "m/s"

    y: SpeedVector = []

    assert type(y) is list
    # 这里没办法判断是SpeedVector类型
    assert type(y) is not SpeedVector


def test_length():
    x = Length(quantity=2, unit="cm")
    y = Length.parse_str("10|cm")

    z1 = x + y
    assert z1.quantity == 12
    z2 = x - y
    assert z2.quantity == -8
    print(z1, z2)


def test_merge_value():
    x = Length.parse_tuple([10, "m"])
    y = Length.parse_tuple([2.1, "cm"])
    z = Length.parse_tuple([20, "cm"])
    try:
        merge_quantified_value([x, y])
    except Exception as e:
        print(e)

    a = merge_quantified_value([x, y, z])
    assert a.quantity == 10.221
    assert a.unit == "m"

    a = merge_quantified_value([x, y, z], target_unit="cm")
    assert a.quantity == 1022.1
    assert a.unit == "cm"


def test_compare():

    a = QuantifiedValue(quantity=10)
    b = Capacity(quantity=11)
    assert a < b
    assert b > a


def test_parse():
    v = QuantifiedValue.model_validate({"quantity": 10, "unit": "m"})
    assert v.quantity == 10
    assert v.unit == "m"


def test_deserialize():
    T = TypeVar("T")
    from logis.data_type import DEFAULT_PYDANTIC_MODEL_CONFIG

    class MyEvent(BaseModel):
        event_type: str
        timestamp: Optional[float] = None

    class MyGenericEvent(MyEvent, Generic[T]):

        data: T

        model_config = DEFAULT_PYDANTIC_MODEL_CONFIG

    class InnerData(BaseModel):
        a: int
        b: int

    json_strings = [
        json.dumps({"event_type": "test", "timestamp": i, "data": {"a": i, "b": i**2}})
        for i in range(2, 100000)
    ]

    with pytest.raises(ValueError):
        MyGenericEvent[bytes].model_validate_json(json_strings[0])
        start = time.time()
        for json_string in json_strings:
            m = MyGenericEvent[bytes].model_validate_json(json_string)
            assert isinstance(m.data, bytes)
        dt = time.time() - start
        print(f"bytes deserialize time: {dt}")

    start = time.time()
    # for json_string in json_strings:
    #     m = MyGenericEvent[Dict[Any, Any]].model_validate_json(json_string)
    #     assert isinstance(m.data, dict)
    dt = time.time() - start
    print(f"dict deserialize time: {dt}")

    start = time.time()
    # for json_string in json_strings:
    #     m = MyGenericEvent[InnerData].model_validate_json(json_string)
    #     assert isinstance(m.data, InnerData)
    dt = time.time() - start
    print(f"inner data deserialize time: {dt}")

    start = time.time()
    for json_string in json_strings:
        m = MyEvent.model_validate_json(json_string)
        assert m
    dt = time.time() - start
    print(f"inner data deserialize time: {dt}")



from typing import Any, Optional, Protocol, Tuple, Union

from pydantic import BaseModel

from logis.data_type import (
    DEFAULT_PYDANTIC_MODEL_CONFIG,
    Capacity,
    Length,
    Point,
    Speed,
    SpeedVector,
    ThreeDimensionalVelocity,
    Time,
    get_time,
)


class TransportProperties(BaseModel):
    """
    搬运相关属性
    """

    # 单次最大搬运量
    single_max_capacity: Optional[Capacity] = None

    # 起始位置
    start_location: Optional[Point] = None
    # 当前位置
    current_location: Optional[Point] = None
    # 终点位置
    target_location: Optional[Point] = None

    @property
    def distance_vector(self) -> Optional[Point]:
        """
        获取设备位置到终点位置的向量
        """
        start = self.current_location
        end = self.target_location
        if start is None or end is None:
            return None
        return end - start

    # 移动速度
    speed: Union[Speed, ThreeDimensionalVelocity, None] = None

    # 装载速度
    load_speed: Optional[Speed] = None
    # 卸载速度,如果未指定，就使用装载速度
    unload_speed: Optional[Speed] = None

    # 装载距离
    load_distance: Optional[Length] = None
    # 卸载距离，如果未指定，使用装载距离
    unload_distance: Optional[Length] = None

    # 装载时间
    load_time: Optional[Time] = None
    # 卸载时间
    unload_time: Optional[Time] = None

    payload: Any = None

    model_config = DEFAULT_PYDANTIC_MODEL_CONFIG

    @property
    def x_speed(self):
        return self.speed.x

    @property
    def y_speed(self):
        return self.speed.y

    @property
    def z_speed(self):
        return self.speed.z

    def get_load_time(self) -> Time:
        """
        获取装载时间
        """
        return (
            self.load_time
            if self.load_time
            else get_time(self.load_distance, self.load_speed)
        )

    def get_unload_time(self) -> Time:
        """
        获取卸载时间
        """
        return (
            self.unload_time
            if self.unload_time
            else get_time(self.unload_distance, self.unload_speed)
        )


class StackerProperties(TransportProperties):
    """
    堆垛机属性
    """

    # 横向长度
    width: Optional[Length] = None

    # 纵向高度
    height: Optional[Length] = None

    # 总容量
    capacity: Optional[Capacity] = None

    # 已用容量
    used_capacity: Optional[Capacity] = None



from enum import Enum
from typing import Literal, NewType, TypeAlias, Union

default_model_name = "default"


class ComponentType(Enum):
    """
    组件类型枚举
    """

    RectangularNode = (
        60003,
        "矩形节点",
    )
    PalletRack = (
        60005,
        "托盘货架",
    )
    AgvRack = (
        60006,
        "AGV货架",
    )
    Asrs = (
        60018,
        "立库",
    )
    PickingRack = (
        60014,
        "挑拣货架",
    )
    AgvAgent = (
        70001,
        "AGV智能体",
    )
    ForkliftAgent = (
        70002,
        "叉车智能体",
    )
    StockAgent = (
        70003,
        "货物智能体",
    )

    def label(self):
        return self.value[1]

    def id(self):
        return self.value[0]

    def id_str(self) -> str:
        return str(self.id())

    def has_id(self, id: Union[str, int]):
        """
        以非严格模式判断id是否相等
        """
        return self.id_str() == str(id)

    def to_dict(self):
        return dict(id=self.id(), label=self.label(), id_str=self.id_str())


class OperationType(Enum):
    """
    操作类型
    """

    Pick = "pick"
    Store = "store"

    def matches(self, operation: Union[str, "OperationType"]):
        return self == operation or self.value == operation


LegencyStorageSelectionStrategyName: TypeAlias = Literal[
    "按储位编号从大到小",
    "按储位编号从小到大",
    "按作业数量少的货架优先",
    "按距离近的货架优先",
]


class StorageSelectionStrategy(Enum):
    """
    存储选择策略枚举，此枚举试图统一并兼容历史逻辑
    """

    BusyLevelAscend = "按作业数量少的优先"
    DistanceAscend = "按距离近的优先"
    NumberAscend = "按编号从小到大"
    NumberDescend = "按编号从大到小"
    Custom = "自定义"

    def matches(
        self,
        strategy: Union[
            "StorageSelectionStrategy", LegencyStorageSelectionStrategyName
        ],
    ):
        # 下面有一些历史兼容逻辑
        if strategy == "按储位编号从大到小":
            return self == StorageSelectionStrategy.NumberDescend
        if strategy == "按储位编号从小到大":
            return self == StorageSelectionStrategy.NumberAscend
        if strategy == "按作业数量少的货架优先":
            return self == StorageSelectionStrategy.BusyLevelAscend
        if strategy == "按距离近的货架优先":
            return self == StorageSelectionStrategy.DistanceAscend
        return self == strategy or self.value == strategy


AgentSelectionStrategyName: TypeAlias = Literal["距离近优先", "空闲优先", "自定义"]
AgentIdleStrategyOption: TypeAlias = Literal["返回到归属地位置", "停留在原地", "自定义"]
GoHomeStrategyFrequency: TypeAlias = Literal["每次", "如果无其他任务"]


TaskType = NewType("TaskType", str)
BatchIn = TaskType("整批入库")
BatchOut = TaskType("整批出库")
ScatterIn = TaskType("零散入库")
ScatterOut = TaskType("零散出库")

TriggerMode = NewType("TriggerMode", str)
ByOrder = TriggerMode("按订单表")
ByEvent = TriggerMode("按事件")

RackType = NewType("RackType", str)

ReplenishmentType = Literal["库区补货", "储位补货"]

LOGIC_TOPO_FILE_EXT = ".logic.gexf"
SCENE_TOPO_FILE_EXT = ".scene.gexf"
TASK_GRAPH_FILE_EXT = ".task_graph.gexf"
PRODUCE_RECIPE_GRAPH_FILE_EXT = ".produce_recipe_graph.gexf"
RACK_GRAPH_FILE_EXT = ".rack_graph.gexf"
NETWORK_FILE_EXT = ".network.gexf"
SQLITE_EXT = ".db"



from diagrams import Diagram
from diagrams.aws.compute import EC2
from diagrams.aws.database import RDS
from diagrams.aws.network import ELB
from diagrams.programming.language import Python

if __name__ == "__main__":
    with Diagram("Web Service", filename="./data/test", show=True):
        ELB("lb") >> EC2("web") >> RDS("userdb") >> Python("xxx")



from typing import Any, Dict, List, Literal, Optional, Union

from logis.data_type import Point

Grid2 = List[List[Point]]
Grid3 = List[Grid2]

from .dict_util import get_the_first_existent_key


def to_grid(
    points: List[Point], include_x=True, include_y=True, include_z=False, **kwargs
) -> Union[Grid3, Grid2, None]:
    """
    将一系列点转换为网格结构
    每个Point代表一个方块的中心点，方块大小一致但不一定是正立方体

    Args:
        points: 一系列Point对象
        **kwargs: 可选参数
            x_step: x方向上的网格步长,如果未指定则自动推测
            y_step: y方向上的网格步长,如果未指定则自动推测
            z_step: z方向上的网格步长,如果未指定则自动推测

    Returns:
        三维网格/二维网格
    """
    if not points:
        return None

    # 定义一个函数来计算步长
    def calculate_step(values):
        """计算一组值的最小步长"""
        if len(values) <= 1:
            return 1

        # 计算相邻值之间的差值
        diffs = [abs(values[i + 1] - values[i]) for i in range(len(values) - 1)]
        return min(diffs)

    x_step = y_step = z_step = None
    x_size = y_size = z_size = None

    # 获取步长参数，如果未指定则自动计算
    if include_x:
        x_step = kwargs.get("x_step", calculate_step(sorted(set(p.x for p in points))))
        min_x = min(p.x for p in points)
        max_x = max(p.x for p in points)
        x_size = int((max_x - min_x) / x_step) + 1

    if include_y:
        y_step = kwargs.get("y_step", calculate_step(sorted(set(p.y for p in points))))
        min_y = min(p.y for p in points)
        max_y = max(p.y for p in points)
        y_size = int((max_y - min_y) / y_step) + 1

    if include_z:
        z_step = kwargs.get("z_step", calculate_step(sorted(set(p.z for p in points))))
        min_z = min(p.z for p in points)
        max_z = max(p.z for p in points)
        z_size = int((max_z - min_z) / z_step) + 1

    valid_indexes = list(filter(lambda v: v is not None, [x_size, y_size, z_size]))
    if len(valid_indexes) < 2:
        return None

    # 初始化网格
    if x_size and y_size and z_size:
        grid = [
            [[None for _ in range(z_size)] for _ in range(y_size)]
            for _ in range(x_size)
        ]
    else:
        x_size, y_size = valid_indexes
        grid = [[None for _ in range(y_size)] for _ in range(x_size)]

    # 将点放入网格中
    for point in points:
        # 计算点在网格中的索引
        x_idx = int((point.x - min_x) / x_step) if x_step else None
        y_idx = int((point.y - min_y) / y_step) if y_step else None
        z_idx = int((point.z - min_z) / z_step) if z_step else None

        tmp = grid
        for _, idx in enumerate([x_idx, y_idx, z_idx]):
            if idx is None:
                continue
            if tmp[idx] is None:
                tmp[idx] = point
            else:
                tmp = tmp[idx]
    return grid


import networkx as nx
from pyecharts import options as opts
from pyecharts.charts import Graph, Sankey, Tree


def nx_digraph_to_graph(
    G: nx.DiGraph,
    title: Optional[str] = None,
    node_size: int = 30,
    node_size_key: str = None,
    layout: Literal["force", "circular"] = "force",
    graph_opts: Dict[str, Any] = {},
    virtual_root_id: Optional[str] = None,
    **kwargs,
) -> Graph:
    """
    将 networkx.DiGraph 转换为 Pyecharts Graph 图表

    参数:
        G: networkx 有向图
        title: 图表标题
        node_size: 默认节点大小
        node_size_key: 从节点属性中读取大小的字段（如节点有 "size" 属性则优先使用）
        edge_width: 默认边宽度
        edge_width_key: 从边属性中读取宽度的字段（如边有 "weight" 属性则优先使用）
        layout: 布局方式

    返回:
        Pyecharts Graph 对象，可直接调用 render() 生成 HTML
    """
    if virtual_root_id and (roots := [n for n, d in G.in_degree() if d == 0]):
        if len(roots) > 1:
            G = G.copy()
            for r in roots:
                G.add_edge(virtual_root_id, r)

    nodes = []
    for node, attrs in G.nodes(data=True):
        size = attrs.get(node_size_key, node_size) if node_size_key else node_size
        nodes.append(opts.GraphNode(name=str(node), symbol_size=size))

    links = []
    for u, v, attrs in G.edges(data=True):
        _, label = get_the_first_existent_key(attrs, "label", "name", "id")
        links.append(
            opts.GraphLink(
                source=str(u),
                target=str(v),
                label_opts=opts.LabelOpts(
                    is_show=False, position="middle", formatter=label, font_size=15
                ),
                emphasis_label_opts=opts.LabelOpts(
                    is_show=True,
                ),
                linestyle_opts=opts.LineStyleOpts(),
            )
        )

    graph = (
        Graph()
        .add(
            series_name="",
            nodes=nodes,
            links=links,
            layout=layout,
            is_draggable=True,
            edge_symbol=graph_opts.pop("edge_symbol", ["circle", "arrow"]),
            **graph_opts,
        )
        .set_global_opts(
            # datazoom_opts=opts.DataZoomOpts(orient="vertical"), # not work
            title_opts=opts.TitleOpts(title=title),
            tooltip_opts=opts.TooltipOpts(trigger="item"),  # 鼠标悬停显示信息
        )
    )

    return graph


def nx_digraph_to_tree(
    G: nx.DiGraph,
    title: Optional[str] = None,
    layout: str = "orthogonal",
    orient: str = "LR",
    name_key: str = "name",
    color_key: str = "color",
    default_name: str = "node",
    virtual_root_id: Optional[str] = None,
    initial_tree_depth: int | None = None,
    tooltip_opts=opts.TooltipOpts(trigger="item", formatter="{b}"),
) -> Tree:
    """
    将树状结构的 networkx.DiGraph 转换为 Pyecharts Tree 图表

    参数:
        G: 有向图（需为无环图且含单根节点）
        title: 图表标题
        layout: 布局方式（"orthogonal" 或 "radial"）
        orient: 树方向（"LR"/"RL"/"TB"/"BT"）
        name_key: 节点属性中用于显示名称的字段
        color_key: 节点属性中用于颜色的字段
        default_name: 无名称属性时的默认名称
        virtual_root_id: 虚拟根节点ID（当节点存在多个根节点时使用）
    """
    # 校验图结构（无环 + 单根节点）
    if not nx.is_directed_acyclic_graph(G):
        raise ValueError("输入图必须是无环有向图（DAG）")
    roots = [n for n, d in G.in_degree() if d == 0]

    if virtual_root_id and len(roots) > 1:
        if virtual_root_id not in G:
            G = G.copy()
            for root in roots:
                G.add_edge(virtual_root_id, root)
        roots = [virtual_root_id]
    elif len(roots) != 1:
        raise ValueError("图中必须有且仅有一个根节点（或指定虚拟根节点）")

    # 递归构建 Tree 数据
    def build_node(current: Any) -> Dict[str, Any]:
        attrs = G.nodes.get(current, {})
        node = {
            "nodeAttrs": attrs,
            "name": attrs.get(name_key, str(current) or default_name),
            "itemStyle": {"color": attrs[color_key]} if color_key in attrs else {},
        }
        children = [build_node(child) for _, child in G.out_edges(current)]
        if children:
            node["children"] = children
        return node

    tree = Tree().set_global_opts(
        # datazoom_opts=opts.DataZoomOpts(orient="vertical"),# not work
        title_opts=opts.TitleOpts(title=title),
        tooltip_opts=tooltip_opts,
    )
    for root in roots:
        tree_data = [build_node(root)]
        tree.add(
            series_name=str(root),
            data=tree_data,
            layout=layout,
            orient=orient,
            label_opts=opts.LabelOpts(
                position="left" if orient in ["LR", "RL"] else "top",
                vertical_align="middle",
                horizontal_align="right",
            ),
            initial_tree_depth=initial_tree_depth,
        )

    return tree


def nx_digraph_to_sankey(
    G: nx.DiGraph,
    title: Optional[str] = None,
    level_key: str = None,
    **kwargs,
) -> Sankey:
    """
    将 networkx.DiGraph 转换为 Pyecharts Sankey 图（适用于有向无环图DAG）

    参数:
        G: 有向图（建议DAG，避免循环影响层级）
        title: 图表标题
        weight_key: 边属性中权重的字段名（用于边的粗细）
        default_weight: 边无权重时的默认值
        level_key: 节点属性中存储层级的字段（如节点有 "level" 属性则优先使用）
                   若为 None，自动通过拓扑排序推断层级
        curve: 边的弯曲度（0为直线，1为最大弯曲）
        node_width: 节点宽度（像素）
        node_gap: 同层级节点间距（像素）
    """
    if level_key is not None:
        # 从节点属性获取层级
        node_levels = {
            node: attrs.get(level_key, 0) for node, attrs in G.nodes(data=True)
        }
    else:
        if not nx.is_directed_acyclic_graph(G):
            raise ValueError(
                "自动推断层级需要输入有向无环图（DAG），或通过 level_key 指定层级"
            )
        topo_order = list(nx.topological_sort(G))
        node_levels = {node: 0 for node in topo_order}
        for node in topo_order:
            for predecessor in G.predecessors(node):
                if node_levels[node] <= node_levels[predecessor]:
                    node_levels[node] = node_levels[predecessor] + 1

    nodes: List[Dict[str, Any]] = []
    for node, attrs in G.nodes(data=True):
        node_name = attrs.get("name", str(node))
        nodes.append(
            {
                "name": node_name,
                "level": node_levels[node],
            }
        )

    links: List[Dict[str, Any]] = []
    for u, v, attrs in G.edges(data=True):
        u_name = G.nodes[u].get("name", str(u))
        v_name = G.nodes[v].get("name", str(v))
        links.append(
            {
                "source": u_name,
                "target": v_name,
                "value": max(1, len(u_name.split(","))) * 5,
            }
        )

    return (
        Sankey()
        .add(
            series_name=title or "Sankey",
            nodes=nodes,
            links=links,
            **kwargs,
        )
        .set_global_opts(
            title_opts=opts.TitleOpts(title=title),
            # datazoom_opts=opts.DataZoomOpts(), # not work
        )
    )


if __name__ == "__main__":
    grid1 = to_grid(
        [Point.of(0, 0, 0), Point.of(1, 1.5, 1), Point.of(2, 2, 2)], include_z=False
    )
    print(grid1)



from decimal import ROUND_HALF_UP, Decimal
from itertools import count
from numbers import Number
from typing import Dict, Generic, Optional, Tuple, TypeVar

from logis.data_type import Number, TuplePoint

from ._math import compute

Num = TypeVar("T", bound=Number)


def round_if_not_none(
    value: Optional[Num], ndigits: Optional[int] = None, rounding=ROUND_HALF_UP
) -> Optional[Num]:
    """
    当value和ndigits都不为None时，进行四舍五入，否则返回原数值。

    :param value: 要四舍五入的值
    :type value: Optional[T]
    :param ndigits: 四舍五入的小数位数，**与原round的行为不同的是，这里为None时返回原数值**
    :type ndigits: Optional[int]
    :return: 四舍五入后的值
    :rtype: T | None
    """
    if value is not None and ndigits is not None:
        # 这里如果不用str转换，float的精度问题仍然会导致结果不正确，例如1216.745、1216.74512等
        v = Decimal(str(value)).quantize(
            Decimal("0." + "0" * ndigits), rounding=rounding
        )
        # if int(value) == 976:
        #     print(value, v, float(v))
        if type(value) is type(v):
            return v
        return type(value)(v)
    return value


def cast_if_not_none(value: Optional[Num], type_: type[Num]) -> Optional[Num]:
    """
    当value不为None时，转换为指定类型，否则返回None。

    :param value: 要转换的值
    :type value: Optional[T]
    :param type_: 目标类型
    :type type_: type[T]
    :return: 转换后的值
    :rtype: T | None
    """
    if value is not None:
        return type_(value)
    return None


class GenericPoint(Generic[Num]):
    __global_precision__: Optional[int] = None

    @property
    def precision(self) -> Optional[int]:
        """
        优先读取__precision_，否则读取__global_precision__
        """
        return self.__precision__ or self.__class__.__global_precision__

    @precision.setter
    def precision(self, value: Optional[int]):
        self.__precision__ = value

    @classmethod
    def model_validate(cls, v: Dict[str, Num]):
        if v is None:
            return None
        if isinstance(v, dict):
            return cls(**v)
        raise ValueError("only dict type is supported")

    def __init__(
        self,
        x: Optional[Num] = None,
        y: Optional[Num] = None,
        z: Optional[Num] = None,
        *args,
        precision: Optional[int] = None,
        **kwargs,
    ):
        """
        Args:
            x (T | None, optional): X coordinate. Defaults to None.
            y (T | None, optional): Y coordinate. Defaults to None.
            z (T | None, optional): Z coordinate. Defaults to None.
            precision (int | None, optional): Precision. Defaults to None.
        """
        self.precision = precision
        if (p := self.precision) is not None:
            assert p > 0, "precision must be greater than 0."
            x = round_if_not_none(x, p)
            y = round_if_not_none(y, p)
            z = round_if_not_none(z, p)
        self.x, self.y, self.z = x, y, z

    @classmethod
    def from_tuple(
        cls, tp: Tuple[Num], precision: Optional[int] = None, **kwargs
    ) -> "GenericPoint[Num]":
        return cls(*tp, precision=precision, **kwargs)

    def __add__(self, other: "GenericPoint[Num]") -> "GenericPoint[Num]":
        assert self.precision == other.precision, "Precision must be the same."
        x = compute("add", self.x, other.x)
        y = compute("add", self.y, other.y)
        z = compute("add", self.z, other.z)
        return self.__class__(x, y, z, precision=self.precision)

    def __sub__(self, other: "GenericPoint[Num]") -> "GenericPoint[Num]":
        assert self.precision == other.precision, "Precision must be the same."
        x = compute("subtract", self.x, other.x)
        y = compute("subtract", self.y, other.y)
        z = compute("subtract", self.z, other.z)
        return self.__class__(x, y, z, precision=self.precision)

    def __mul__(self, other: Number) -> "GenericPoint[Num]":
        assert other is not None, "Multiplier must not be None."
        x = compute("multiply", self.x, other) if self.x is not None else None
        y = compute("multiply", self.y, other) if self.y is not None else None
        z = compute("multiply", self.z, other) if self.z is not None else None
        return self.__class__(x, y, z, precision=self.precision)

    def __truediv__(self, other: Number) -> "GenericPoint[Num]":
        assert other is not None, "Divisor must not be None."
        x = compute("divide", self.x, other) if self.x is not None else None
        y = compute("divide", self.y, other) if self.y is not None else None
        z = compute("divide", self.z, other) if self.z is not None else None
        return self.__class__(x, y, z, precision=self.precision)

    def __eq__(self, value):
        if not isinstance(value, self.__class__):
            return False
        dv = 1 / (10**self.precision) if self.precision else None

        def is_acceptable(a: Optional[Number], b: Optional[Number]) -> bool:
            if dv is None:
                return a == b
            if a is None and b is None:
                return True
            if a is None or b is None:
                return False
            return abs(a - b) < dv

        return (
            is_acceptable(self.x, value.x)
            and is_acceptable(self.y, value.y)
            and is_acceptable(self.z, value.z)
            and self.precision == value.precision
        )


class Point(GenericPoint[float]):
    """
    此类不通用，历史遗留，不建议使用，如有需要建议使用GenericPoint
    支持x,y,z三维坐标点，也可当作二维坐标使用
    TODO: 处理单位
    """

    __counter = count(0, step=1)
    __global_precision__ = 3

    @classmethod
    def try_parse(cls, dc: dict, **kwargs) -> "Point":
        """
        尝试读取X、Y、Z坐标，返回一个Point对象
        """
        if not isinstance(dc, dict):
            raise ValueError("Expected a dictionary for Point parsing.")
        xyz = [dc.get(k, None) for k in ("X", "Y", "Z")]
        return Point(xyz, **kwargs)

    @classmethod
    def from_tuple(cls, args: Tuple[Number], **kwargs):
        """
        从元组创建点
        """
        return cls(args, **kwargs)

    @classmethod
    def of(
        cls,
        x: Optional[Number] = None,
        y: Optional[Number] = None,
        z: Optional[Number] = None,
        **kwargs,
    ) -> "Point":
        p = Point(x, y, z, **kwargs)
        return p

    def __init__(self, *args, precision: Optional[int] = None, **kwargs):
        """
        支持[x,y]、x,y、[x,y,z]、x,y,z传参赋值
        默认x=y=z=0

        TODO: 不应该自动转float、不应该自动保留两位小数
        """
        self._id = next(self.__counter)
        self.unit: Optional[str] = kwargs.get("unit", None)

        x: Optional[Number] = kwargs.get("x", None)
        y: Optional[Number] = kwargs.get("y", None)
        z: Optional[Number] = kwargs.get("z", None)
        if len(args) == 1:
            assert isinstance(
                args[0], (list, tuple)
            ), "single argument must be a list or tuple."
            x = cast_if_not_none(args[0][0], float)
            y = cast_if_not_none(args[0][1], float)
            if len(args[0]) > 2 and (z := args[0][2]) is not None:
                z = cast_if_not_none(z, float)

        elif len(args) >= 2:
            x = cast_if_not_none(args[0], float)
            y = cast_if_not_none(args[1], float)
            if len(args) > 2 and (z := args[2]) is not None:
                z = cast_if_not_none(z, float)

        # 调试过程中发现z有时候是None有时候是0，因此这里统一给了默认值
        # 实际上应该外层规范
        x = x if x is not None else 0
        y = y if y is not None else 0
        z = z if z is not None else 0

        kwargs.update(x=x, y=y, z=z)

        super().__init__(precision=precision, **kwargs)

    def __repr__(self):
        return f"Point({self.x}, {self.y}, {self.z})"

    def __hash__(self):

        def s(v: Optional[Number]):
            if not self.precision:
                return v
            if v is None:
                return v
            return v
            # return str(Decimal(v).quantize(Decimal(f"0.{'0'*self.precision}")))

        return hash((s(self.x), s(self.y), s(self.z)))

    def __lt__(self, other):
        """
        FIXME: 考虑z
        """
        if isinstance(other, Point):
            # TODO: 只要任何一个大于就行？
            return (self.x, self.y) < (other.x, other.y)
        return NotImplemented

    def __sub__(self, other: "Point"):
        def sub(a: Number, b: Number):
            if a is None and b is None:
                return None
            a = a or 0
            b = b or 0
            return a - b

        assert self.unit == other.unit, "unit must be same"
        return Point.of(
            x=sub(self.x, other.x),
            y=sub(self.y, other.y),
            z=sub(self.z, other.z),
            unit=self.unit,
        )

    def to_tuple(self) -> TuplePoint:
        return (
            self.x,
            self.y,
            self.z,
        )

    def __str__(self):
        return f"Point(id={self._id},x={self.x},y={self.y},z={self.z})"


if __name__ == "__main__":
    pass



from abc import abstractmethod
from collections import defaultdict
from typing import Dict, List, Optional, Set

import simpy
from networkx import DiGraph

from logis.data_type import Point
from logis.iface.graph import IPathGraph

from ..model import DirEdge


class ISimPathGraph(IPathGraph[DirEdge, Point]):
    """
    路径关系图
    """

    @property
    @abstractmethod
    def env(self) -> simpy.Environment:
        pass

    def __init__(self, *args, **kwargs) -> None:
        super().__init__()
        # TODO: 是否直接继承会更方便？
        self._graph_ = DiGraph()
        self._lock_nodes_map: Dict[str, Set[Point]] = defaultdict(lambda: set())
        self._path_lock = simpy.Resource(self.env)

    def request_path_lock(self):
        return self._path_lock.request()

    def nodes(self, *args, **kwargs) -> List[Point]:
        return list(self._graph_.nodes(*args, **kwargs))

    def edges(self, *args, **kwargs) -> List[DirEdge]:
        return list(self._graph_.edges(*args, **kwargs))

    def lock_node(self, lock_id: str, node: Point):
        self._lock_nodes_map[lock_id].add(node)

    def unlock_node(self, lock_id: str, node: Point):
        if node in self._lock_nodes_map[lock_id]:
            self._lock_nodes_map[lock_id].remove(node)

    def unlock_nodes_by_lock_id(self, lock_id: str):
        self._lock_nodes_map[lock_id].clear()

    def get_locked_by_other(self, my_lock_id: str):
        return set().union(
            *[nodes for id, nodes in self._lock_nodes_map.items() if id != my_lock_id]
        )

    def lock_path(self, lock_id: str, path: List[Point], limit: Optional[int] = None):
        """
        锁格
        """
        path = path[0:limit]
        while True:
            with self.request_path_lock() as req:
                yield req
                locked = self.get_locked_by_other(lock_id)
                if len(locked.intersection(path)) == 0:
                    # 在获取锁的情况下立即锁定路径
                    for p in path:
                        self._lock_nodes_map[lock_id].add(p)
                    break
            # 未获取锁或路径被锁定，等待后重试
            yield self.env.timeout(0.5)

    def unlock_path(self, lock_id: str, path: List[Point]):
        for p in path:
            if p in self._lock_nodes_map[lock_id]:
                self._lock_nodes_map[lock_id].remove(p)

    def add_nodes(self, *nodes: Point):
        for node in nodes:
            self._graph_.add_node(node)

    def add_edges(self, *edges: DirEdge):
        for edge in edges:
            self._graph_.add_edge(edge.starter, edge.ender, weight=edge.get_weight())

    def remove_edges(self, *edges: DirEdge):
        for edge in edges:
            self._graph_.remove_edge(edge.starter, edge.ender)



from typing import Protocol, runtime_checkable

from logis.biz.sim.stock.model import QuantifiedStock

from ..model import *


@runtime_checkable
class StorageStrategy(Protocol):
    """
    存取策略
    """

    def find_location(
        self, storage: StorageProperties, stocks: List[QuantifiedStock], *args, **kwargs
    ) -> List[CellProperties]:
        """
        寻找储位位置，存储获取或者取出货物
        """
        raise NotImplementedError()



from datetime import datetime, timezone
from pathlib import Path
from random import randint
from typing import Literal

from influxdb_client.client.influxdb_client import InfluxDBClient
from influxdb_client.client.influxdb_client_async import InfluxDBClientAsync
from influxdb_client_3 import InfluxDBClient3, Point

from logis.metric.influxdb import InfluxCommand, InfluxRestClient

database = "test"
test_url = f"http://localhost:48181"
admin_token_file = Path(__file__)
token = "apiv3_9RtOrHrNRybmP8s-VnjxFU5nb89Ly_Xu5iWDN4vFUnPlK4-M0gbE0ODsLh4OGYvUyBGHsBwOm1RpDqQ-ZUqsyg"


def get_client(
    url=test_url,
    token=token,
    mode: Literal["sync", "async"] = "sync",
    database=database,
):
    return InfluxDBClient3(host=url, token=token, database=database)


def test_insert():
    client: InfluxDBClient3 = get_client(mode="sync")
    for i in range(10):
        client.write(
            record=Point("andy.temperature")
            .field("value", randint(36, 40) + 0.5)
            .tag("id", str(i))
            .tag("source", "体温计")
            .time(datetime.now()),
        )
    client.close()


def test_query():
    client: InfluxDBClient3 = get_client(mode="sync")
    result = client.query(
        'select * from "andy.temperature" where id>5 order by time desc'
    )
    print(result.to_pandas())
    client.close()


def test_command():
    bin_dir = Path("D:\\app\\influxdb3-core-3.5.0-windows_amd64")
    command = InfluxCommand(bin_dir)
    token = command.generate_admin_token(
        offline=False, path=bin_dir / "admin_token.json"
    )
    print("Process exited with code:", token)

    token = command.generate_admin_token(
        offline=True, path=bin_dir / "admin_token.json"
    )
    print("Process exited with code:", token)


def test_serve():
    bin_dir = Path("D:\\app\\influxdb3-core-3.5.0-windows_amd64")
    command = InfluxCommand(bin_dir)
    p = command.serve(
        data_dir=bin_dir / "data",
        object_store="file",
    )
    print(p)


def test_rest_client():
    tmp_db = "xxxxxx"
    client = InfluxRestClient("http://127.0.0.1:48181", token=token)
    x = client.create_database(tmp_db)
    assert x.success or x.error.code == 409, x
    x = client.delete_database(tmp_db)
    assert x.success, x



import pytest

from logis.data_type.exception import MyBaseException


def test_try_except():
    a = 0
    with pytest.raises(ZeroDivisionError):
        try:
            1 / 0
        finally:
            a = 1

    assert a == 1


def test_exception():
    e = Exception("test", "xxxx")

    assert e.args == ("test", "xxxx")

    assert str(e) == "('test', 'xxxx')"


def test_my_base_exception():
    e = MyBaseException("test", "xxxx", args=(1, 2, 3), v=1)

    assert e.v == 1
    assert e.args == ("test", "xxxx")

    assert str(e) == "('test', 'xxxx')"



import logging
from collections import defaultdict
from typing import Dict, List, Literal, Optional, Set, Type, Union, get_args, get_origin

from pydantic import AliasChoices, Field, ValidationInfo, field_validator

from logis.biz.sim.const import RackType
from logis.biz.sim.stock import QuantifiedStock
from logis.data_type import (
    DEFAULT_PYDANTIC_MODEL_CONFIG,
    Capacity,
    QuantifiedValue,
    SpatialProps,
    Time,
)


class ContainerMetadata(SpatialProps):
    """
    容器元数据
    """

    model_config = DEFAULT_PYDANTIC_MODEL_CONFIG
    id: Optional[str] = Field(
        validation_alias=AliasChoices("id", "平堆区编号", "货架编号", "储位编码"),
        default=None,
    )

    name: Optional[str] = Field(
        validation_alias=AliasChoices("名称", "name"), default=None
    )

    capacity: Optional[Capacity] = None
    occupied_capacity: Optional[Capacity] = None

    @field_validator("width", "height", "depth", mode="before")
    def check_width(cls: Type[QuantifiedValue], v, validation_info: ValidationInfo):
        """
        其实这里为了避免重复逻辑，就要通过反射来获取字段的类型，麻烦了点但是也长了见识
        """
        field = cls.model_fields.get(validation_info.field_name)
        # field_type = get_origin(field.annotation)
        field_types = get_args(field.annotation)
        field_types: List[Type[QuantifiedValue]] = list(
            filter(lambda t: issubclass(t, QuantifiedValue), field_types)
        )
        if isinstance(v, str) and field_types:
            v = field_types[0].parse_str(v)
        return v

    @field_validator("rotation", mode="before")
    def check_rotation(cls, v, validation_info: ValidationInfo):
        if not v:
            return None
        try:
            return QuantifiedValue(quantity=float(v), unit=None)
        except Exception as e:
            logging.warning("convert failed：%s", e)
        return None

    @property
    def free_capacity(self) -> Capacity:
        return (
            self.capacity - self.occupied_capacity
            if self.occupied_capacity
            else self.capacity
        )

    disabled: Optional[bool] = False

    store_speed_time: Optional[Time] = None
    # 如果不指定默认使用store_speed_time
    retrieve_speed_time: Optional[Time] = None

    # # 是否是互斥访问
    # exclusive: bool = True

    # 内部可以放置的物品类型
    allowed_stock_type: Optional[Union[str, List[str]]] = None

    @property
    def allowed_stock_types(self):
        """
        这里为了方便访问，返回一个固定格式的列表
        """
        if isinstance(self.allowed_stock_type, str):
            return [self.allowed_stock_type]
        elif isinstance(self.allowed_stock_type, list):
            return self.allowed_stock_type
        return ["*"]

    def get_cells(self) -> List["CellProperties"]:
        """
        获取所有的储位
        """
        if hasattr(self, "cells"):
            return self.cells
        return [self]

    def group_cell_by_stock_type(self) -> Dict[str, List["CellProperties"]]:
        dc = defaultdict(list)
        for cell in self.get_cells():
            for allowed_type in cell.allowed_stock_types:
                items = dc[allowed_type]
                items.append(cell)
                dc[allowed_type] = items
        return dc

    def is_allowed_stock_type(self, stock_type: str) -> bool:
        tps = self.allowed_stock_types
        return "*" in tps or stock_type in tps

    def can_store(self, v: QuantifiedStock) -> bool:
        """
        判断当前存储单元是否可以存储指定货物
        1. 类型符合
        2. 空间足够
        """
        return self.is_allowed_stock_type(v.unique_id) and self.free_capacity >= v

    def can_retrieve(self, v: QuantifiedStock) -> bool:
        """
        判断当前存储单元是否可以提供指定货物
        1. 类型符合
        2. 空间足够
        3. 未被占用
        """
        if self.occupied_capacity is None:
            return False
        return self.is_allowed_stock_type(v.unique_id) and self.occupied_capacity >= v


class CellProperties(ContainerMetadata):
    """
    储位属性
    """

    model_config = DEFAULT_PYDANTIC_MODEL_CONFIG

    # # 储位编码
    # cell_id: Optional[str] = None
    # 货架编码
    rack_id: Optional[str] = None
    # 货架组编码
    rack_group_id: Optional[str] = None

    # 储位存储的货物信息
    stocks: List["QuantifiedStock"] = []

    def store(self, v: QuantifiedStock) -> bool:
        """
        存储货物
        """
        ok = self.can_store(v)
        if ok:
            self.stocks.append(v)
        return ok

    def retrieve(self, v: QuantifiedStock) -> bool:
        """
        取出货物
        """
        ok = self.can_retrieve(v)
        if ok:
            self.stocks.remove(v)
        return ok


class RackProperties(ContainerMetadata):
    """
    货架的属性信息
    """

    rack_group_id: Optional[str] = None

    type: Optional[RackType] = Field(
        None, validation_alias=AliasChoices("类型", "type")
    )
    # TODO: 考虑去除
    is_obstacle: Optional[bool] = Field(
        validation_alias=AliasChoices("是否障碍", "is_obstacle", "是障碍"), default=None
    )

    row_count: int = Field(
        validation_alias=AliasChoices("层数", "row_count"), default=1
    )
    col_count: Optional[int] = Field(
        validation_alias=AliasChoices("单元格数", "列数", "col_count"), default=None
    )

    cell_metadata: Optional[ContainerMetadata] = None

    cell_ids: List[str] = Field(default_factory=list)
    cells: List[CellProperties] = Field(default_factory=list)


class RackGroupProperties(ContainerMetadata):
    """
    货架组
    """

    model_config = DEFAULT_PYDANTIC_MODEL_CONFIG

    rack_ids: List[str] = Field(default_factory=list)
    racks: List[RackProperties] = Field(default_factory=list)


class AsrsProperties(RackProperties):
    """
    立库属性信息
    """

    type: Literal["一货架，一堆垛机", "两货架，一堆垛机"] = Field(alias="立库系统类型")


StorageProperties = Union[RackGroupProperties, RackProperties, CellProperties]



from abc import ABC, abstractmethod
from ast import Not
from enum import Enum

from logis.data_type import ABCEnumMeta
from logis.iface.handler import IHandler
from logis.task.model import ITask


class IWorkingMode(ABC):
    """
    工作模式
    """

    pass


class WorkingMode(IWorkingMode, Enum, metaclass=ABCEnumMeta):
    """
    工作模式
    """

    SYNC = "synchronous"
    ASYNC = "asynchronous"


class ITaskHandler(IHandler):
    """
    任务处理接口
    """

    def get_current_task_concurrency(self, *args, **kwargs):
        """
        获取当前正在处理的任务并发数
        """
        raise NotImplementedError("get_current_task_concurrency")

    def exceed_the_maximum_task_concurrency(self, *args, **kwargs):
        """
        判断是否达到了任务并发上限
        """
        raise NotImplementedError("exceed_the_maximum_task_concurrency")

    def get_working_mode(self):
        """
        获取处理模式
        """
        raise NotImplementedError("get_working_mode")

    def get_task_manifest(self):
        """
        获取任务清单
        """
        raise NotImplementedError("get_task_manifest")

    def is_task_all_done(self, **kwargs) -> bool:
        """
        任务是否全部完成
        """
        raise NotImplementedError("is_task_all_done")

    @abstractmethod
    def on_task_received(self, *tasks: ITask, **kwargs):
        """
        接收任务
        """

    @abstractmethod
    def on_task_succeeded(self, *args, **kwargs):
        """
        任务完成
        """

    def on_task_failed(self, *args, **kwargs):
        """
        任务失败
        """
        raise NotImplementedError("on_task_failed")



__doc__ = """
字典工具类
"""

from typing import Any, Dict, Iterable, Optional, Tuple, TypeVar, Union

cn_en_dict = dict(
    半径="radius",
)


def unify_key_name(input: dict, mappings: dict = cn_en_dict) -> dict:
    """
    将字典的键名按照字典中的映射规则进行统一转换
    Args:
        input (dict): 输入的字典，键名可能是中文或其他形式。
        mappings (dict): 键名映射规则，键为原始键名，值为目标键名。
    """
    output = dict()
    for k, v in input.items():
        new_key = mappings.get(k, k)
        if new_key in output:
            raise ValueError(f"Duplicate key found: {new_key} in input dictionary.")
        output[new_key] = v
    return output


def remove_key_not_in(dc: dict, keys: Iterable[Any], delete_value: bool = False):
    """
    移除不在指定范围内的key

    Args:
        delete_value: 是否同时del value
    """
    keys_removed = []
    for key in list(dc.keys()):
        if key not in keys:
            v = dc.pop(key, None)
            if delete_value and v is not None:
                del v
                keys_removed.append(key)
    return keys_removed


K = TypeVar("K")
V = TypeVar("V")


def get_the_first_existent_key(
    obj: Dict[K, V], *keys: K, default: Optional[V] = None
) -> Tuple[Optional[K], Optional[V]]:
    """
    获取第一个存在的key的值

    Args:
        obj (Dict[K,V]):  输入的字典，键名可能是中文或其他形式。
        keys (Iterable[K]): 要查找的键名列表。
        default (V, optional): 如果没有找到任何键，返回的默认值。默认值为None。

    Returns:
        (Tuple[K, V]): 第一个存在的键值对，键为存在的键名，值为对应的值。如果没有找到任何键，返回(None, default)。
    """
    for key in keys:
        if key in obj:
            return key, obj[key]
    return None, default


from collections import Counter


def compare_counter_progress(progress: Counter, target: Counter) -> int:
    """
    比较计数器的进度
    主要是兼容低版本，高版本的Counter可以直接比较
    """
    p_keys = set(progress.keys())
    t_keys = set(target.keys())
    assert p_keys.issubset(t_keys), "progress must be a subset of target"
    dk = len(p_keys) - len(target.keys())
    if dk != 0:
        return dk
    dvs = []
    for k in p_keys:
        v1 = progress.get(k)
        v2 = target.get(k)
        dv = v1 - v2 if v1 != v2 else 0
        if dv > 0:
            dv = 1
        elif dv < 0:
            dv = -1
        if dv != 0:
            dvs.append(dv)

    if len(dvs) == 0:
        return 0

    dv_sum = sum(dvs)

    # 全部相等
    if dv_sum == 0:
        return 0
    # 全部大
    elif dv_sum == len(dvs):
        return 1
    # 全部小
    elif dv_sum == -len(dvs):
        return -1
    # 有的大有的小就是小
    return -1



import inspect
import logging
from importlib import import_module
from types import ModuleType
from typing import Callable, List, Optional, Set, Tuple, Type, TypeVar

import hmr
from watchfiles import Change, PythonFilter, awatch


def try_import(path: str):
    """
    支持以x.y.z的方式导入模块或模块内的属性
    Args:
        path: 模块路径或属性路径
    Returns:
        导入的模块或属性, 如果导入失败则返回None
    """
    try:
        return import_module(path)
    except ModuleNotFoundError:
        tmps = path.split(".")
        assert len(tmps) > 1, f"unexpected path: {path}"
        module = import_module(".".join(tmps[0:-1]))
        return getattr(module, tmps[-1])
    except Exception as e:
        logging.warning("failed to import %s: %s", path, e)
        return None


def try_hot_reload(name, package: Optional[str] = None):
    """
    热加载指定模块
    Args:
        name: 模块名
        package: 包名
    """
    try:
        hmlib = import_module(name=name, package=package)
        logging.info("shall reload: %s", hmlib)
        hmr.reload(hmlib)
    except Exception as e:
        logging.warning("failed to reload: %s", e)


async def watch_python_dir(
    dir: str, on_change: Callable[[Set[Tuple[Change, str]]], None]
):
    """
    这是一个使用样例，功能并不通用
    """
    logging.info("watch files in %s", dir)
    async for change in awatch(str(dir), watch_filter=PythonFilter()):
        logging.info("change detected:%s", change)
        on_change(change)


T = TypeVar("T")


def collect_subclass_of(
    base_class: Type[T], _from: ModuleType, include_base_class: bool = False
):
    def predicate(m):
        if not include_base_class and m is base_class:
            return False
        return inspect.isclass(m) and issubclass(m, base_class)

    name_class_map: List[Tuple[str, T]] = inspect.getmembers(_from, predicate=predicate)
    return name_class_map


def get_class_full_path(cls) -> str:
    """
    获取类的全路径表示（模块路径.类名）

    参数:
        cls: 目标类（需传入类本身，而非实例）

    返回:
        str: 类的全路径，格式为 "模块路径.类名"
    """
    # __qualname__ 获取类的限定名（含嵌套关系，如 "OuterClass.InnerClass"）
    return f"{cls.__module__}.{cls.__qualname__}"



from abc import ABC, abstractmethod

import simpy
import simpy.resources.resource
from ipa.decorator import deprecated


class ISimLock(ABC):

    @property
    @abstractmethod
    def env(self) -> simpy.Environment:
        pass

    def __init__(self, **kwargs):
        self._occupied_ = simpy.Resource(self.env, capacity=1)

    @property
    @deprecated("use is_locked instead")
    def is_occupied(self):
        """
        判断锁是否被占用
        """
        return self.is_locked

    @property
    def is_locked(self):
        return self._occupied_.count > 0

    def lock(self):
        """
        请求锁
        """
        return self._occupied_.request()

    def unlock(self, req: simpy.resources.resource.Request):
        """
        请求解锁
        """
        return self._occupied_.release(req)



import pytest

from logis.iface import Storable


class MyBox(Storable):

    def store(self, item, **config):
        print(item, config)
        return True


class MyBox2:
    def store(self, item, a=2, **config):
        print(item, config)
        return True

    def retrieve(self, item, **config):
        print(item, config)
        return True

    def pre_store(self, item, **config):

        print(item, config)
        return True

    def pre_retrieve(self, item, **config):
        print(item, config)
        return True


class MyBox3:
    def store(self, *args, **config):
        print(args, config)
        return True

    def retrieve(self, *args, **config):
        print(args, config)
        return True

    def pre_store(self, **config):
        pass

    def pre_retrieve(self, **config):
        pass


def test_issubclass():
    assert issubclass(MyBox, Storable)
    x = MyBox()
    assert x.store("hello", xxx=1)

    assert issubclass(MyBox2, Storable)

    assert isinstance(MyBox2(), Storable)
    assert isinstance(MyBox3(), Storable)



import logging
from collections import Counter
from typing import Any, Dict, Generic, List, Optional, TypeVar

from pydantic import BaseModel, Field
from toolz import unique

from logis.biz.sim.stock import IStock
from logis.data_type import DEFAULT_PYDANTIC_MODEL_CONFIG, ComponentId
from logis.task import ITask, TaskId

T = TypeVar("T", bound=ITask)


class TaskManifest(BaseModel, Generic[T]):
    """
    任务清单
    """

    model_config = DEFAULT_PYDANTIC_MODEL_CONFIG

    component_id: Optional[ComponentId] = None

    order_tasks: Dict[TaskId, T] = Field(default_factory=dict)

    stock_tasks: List[IStock] = Field(default_factory=list)

    target_stock_counter: Dict[Any, float] = Field(default_factory=Counter)
    stock_counter: Dict[Any, float] = Field(default_factory=Counter)

    def get_task(self, task_id: TaskId) -> Optional[T]:
        """
        获取任务
        """
        return self.order_tasks.get(task_id, None)

    def add_stock_task(self, *stocks: IStock):
        """
        添加库存到清单中
        """
        if not stocks:
            return
        self.stock_tasks.extend(stocks)

    def remove_stock_task(self, *stocks: IStock):
        """
        从清单中移除库存
        """
        if not stocks:
            return
        for stock in stocks:
            if not stock:
                continue
            if stock in self.stock_tasks:
                self.stock_tasks.remove(stock)
            else:
                logging.error(f"库存{stock}不在清单中")

    @property
    def task_ids(self) -> List[TaskId]:
        """
        获取所有任务ID
        """
        return unique([*self.order_tasks.keys()])

    def add_order_task(self, *order_tasks: T):
        """
        通过判断任务ID是否已存在来判断是否为新任务

        Returns:
            List[OrderTask]: 新添加的订单任务列表
        """
        new_tasks: List[T] = []
        for order_task in order_tasks:
            if not order_task:
                continue
            if order_task.get_task_id() in self.order_tasks:
                continue
            new_tasks.append(order_task)
            self.order_tasks[order_task.get_task_id()] = order_task
        return new_tasks

    def increase_stock_counter(self, stock: "IStock"):
        """
        当前值
        """
        key = (stock.name, stock.unit)
        self.stock_counter[key] += stock.quantity

    def increase_target_stock_counter(self, stock: "IStock"):
        """
        目标值
        """
        key = (stock.name, stock.unit)
        self.target_stock_counter[key] += stock.quantity

    def is_stock_io_balanced(self) -> bool:
        """
        检查出入货物是否平衡
        """
        return self.stock_counter == self.target_stock_counter

    def is_finished(self) -> bool:
        """
        检查任务是否完成
        """
        return self.stock_counter >= self.target_stock_counter

    def no_stock_task_left(self):
        """
        任务队列是否为空
        """
        return not self.stock_tasks



from decimal import Decimal

import pytest

from logis.data_type import GenericPoint, Point


def test_point():
    p1 = Point(1, 2, 0)
    p2 = Point(1, 2, 3)
    p3 = Point.from_tuple([1, 2, 3])
    p4 = Point.from_tuple([10, 1, 0])
    for p in [p1, p2, p3, p4]:
        print(p)

    assert p1 - p2 == Point(0, 0, -3)
    assert p3 - p4 == Point(-9, 1, 3)
    assert Point() - Point() == Point()


def test_generic_point():
    p = GenericPoint[float](1.23456, 2.34567, 3.45678, precision=4)
    assert p.x == 1.2346
    assert p.y == 2.3457
    assert p.z == 3.4568

    p = GenericPoint[Decimal](
        Decimal("1.23456"), Decimal("2.34567"), Decimal("3.45678"), precision=4
    )

    assert p.x == Decimal("1.2346")
    assert p.y == Decimal("2.3457")
    assert p.z == Decimal("3.4568")


def test_compute():
    p1 = GenericPoint[int].from_tuple((1, 2, 3), precision=1)
    p2 = GenericPoint[float](1, 3, 4, precision=1)
    a = p1 + p2
    assert a.x == 2
    assert a.y == 5
    assert a.z == 7

    p3 = GenericPoint[Decimal](
        Decimal("1.14"), Decimal("2.16"), Decimal("3.55"), precision=1
    )
    b = p1 + p3
    assert b.x == Decimal("2.1")
    assert b.y == Decimal("4.2")
    assert b.z == Decimal("6.6")

    c = p2 * 5
    assert c.x == Decimal("5.0")
    assert c.y == Decimal("15.0")
    assert c.z == Decimal("20.0")

    p4 = GenericPoint[int](5, 10)
    d = p4 / 5
    assert d.x == 1
    assert d.y == 2
    assert d.z is None

    p5 = GenericPoint[int](10, 20, 1)
    p6 = GenericPoint[int](10, 10, None)

    with pytest.raises(ValueError):
        p5 / p6



import pytest
from networkx import NetworkXError

from logis.task import QuantifiedTask, Task, TaskGraph, TaskStatus
from logis.task.model import TaskId, TaskPriority


def test_quantified_task():
    class MyTask(QuantifiedTask):
        id: str
        priority: TaskPriority = 0

        def get_task_id(self) -> TaskId:
            return self.id

        def get_priority(self) -> TaskPriority:
            return self.priority

        def is_status_at(self, status):
            return super().is_status_at(status)

        def update_status(self, status: TaskStatus):
            super().update_status(status)

        def model_dump(self, *args, **kwargs):
            return super().model_dump(*args, **kwargs)

        @classmethod
        def model_validate(cls, data, *args, **kwargs):
            pass

    task = MyTask(id="task1", quantity=10)
    assert task.quantity == 10


def test_task_graph():
    graph = TaskGraph()
    task1 = Task(id="task1")
    task2 = Task(id="task2")
    task3 = Task(id="task3")

    graph.add_task(task1)
    graph.add_task_if_absent(task1.model_copy(update=dict(name="task1 copy")))
    assert graph.get_task(task1.id).name is None

    # 测试父子关系
    graph.add_task(task2, parent=task1)
    graph.add_task(task3, parent=task2)
    assert list(graph.get_parents_id(task2.id)) == [task1.id]
    assert list(graph.get_children_id(task1.id)) == [task2.id]
    assert graph.has_parent_child_relationship(task1, task2)
    assert graph.has_parent_child_relationship(task1, task3, strict=True) is False
    assert graph.has_parent_child_relationship(task1, task3, strict=False) is True

    # 测试任务完成状态
    assert graph.is_task_finished(task1) is False
    graph.update_task_status(task3, TaskStatus.FINISHED)
    assert graph.is_task_finished(task1) is True

    # 测试删除
    with pytest.raises(NetworkXError):
        graph.remove_task("not exist")
    graph.remove_task_if_present(task2)
    assert task2 not in graph

    assert list(graph.get_children_id(task1)) == []
    assert task1.get_task_id() in graph

    task4 = Task(id="task4")
    task5 = Task(id="task5")
    graph.add_task(task4, parent=task3)
    graph.add_task(task5, parent=task4)
    graph.remove_task(task3, cascade=True)

    assert graph.task_size() == 1
    assert list(graph.tasks()) == [task1]
    assert list(graph.tasks(only_id=True)) == [task1.id]


def test_task_handler():
    from logis.task import ITaskHandler, IWorkingMode, WorkingMode

    class TestTaskHandler(ITaskHandler):
        def get_working_mode(self) -> IWorkingMode:
            return WorkingMode.SYNC

        def on_task_received(self, *tasks, **kwargs):
            pass

        def handle(self, *args, **kwargs):
            print(args, kwargs)

        def on_task_succeeded(self, *args, **kwargs):
            pass

    handler = TestTaskHandler()

    handler.handle(a=1, b=2, c=3)



from abc import ABCMeta, abstractmethod
from collections import defaultdict
from typing import Iterable, Iterator, Optional

from networkx import DiGraph, has_path

from .model import ITask, TaskId, TaskLike, TaskStatus


class AbstractTaskManager(metaclass=ABCMeta):
    """
    任务管理器基础抽象类
    """

    @abstractmethod
    def is_task_finished(self, task: TaskLike, update: bool = True, infer=True) -> bool:
        """
        检查任务是否已完成
        Args:
            task (TaskLike): 任务对象
            update (bool, optional): 是否在检测到任务已完成时自动更新任务属性
            infer (bool, optional): 是否根据子任务的状态推测本任务的状态
        Returns:
            bool: 任务是否已完成
        """
        pass

    def is_status_at(
        self,
        task: TaskLike,
        status: TaskStatus,
        infer: bool = True,
        update: bool = True,
    ) -> bool:
        """
        检查任务是否在指定状态
        Args:
            task (TaskLike): 任务对象
            status (TaskStatus): 状态
            infer (bool, optional): 是否根据子任务的状态推测本任务的状态，默认值为True
            update (bool, optional): 是否在检测到任务状态符合要求时自动更新任务属性，默认值为True
        Returns:
            bool: 任务是否在指定状态
        """
        raise NotImplementedError()

    @abstractmethod
    def task_size(self) -> int:
        """
        获取任务数量
        """
        pass

    @abstractmethod
    def tasks(self, only_id: bool = False) -> Iterator[ITask]:
        """
        获取所有任务
        """
        pass

    @abstractmethod
    def add_task(self, task: TaskLike, parent: Optional[TaskLike] = None, **attr):
        """
        添加任务

        Args:
            task (TaskLike): 任务对象
            parent (TaskLike | None): 父任务对象. Defaults to None.
        """
        pass

    @abstractmethod
    def add_task_if_absent(
        self, task: TaskLike, parent: Optional[TaskLike] = None, **attr
    ) -> bool:
        """
        如果任务不存在则添加,否则不添加

        Args:
            task (TaskLike): 任务对象
            parent (TaskLike | None): 父任务对象. Defaults to None.

        Returns:
            bool: 是否添加成功
        """
        pass

    @abstractmethod
    def remove_task(self, task: TaskLike, cascade: bool = False):
        """
        删除任务

        Args:
            task (TaskLike): 任务对象
            cascade (bool, optional): 是否级联删除子任务. Defaults to False.
        """
        pass

    @abstractmethod
    def remove_task_if_present(self, task: TaskLike, cascade: bool = False):
        """
        如果任务存在则删除

        Args:
            task (TaskLike): 任务对象
            cascade (bool, optional): 是否级联删除子任务. Defaults to False.
        """
        pass

    @abstractmethod
    def get_task(self, task: TaskLike) -> Optional[ITask]:
        """
        根据任务ID获取任务本身
        """
        pass

    @abstractmethod
    def get_children_id(self, task: TaskLike, strict: bool = True) -> Iterable[TaskId]:
        """
        获取任务的所有子任务ID

        Args:
            task (TaskLike): 任务对象
            strict (bool, optional): 是否严格模式. Defaults to True.

        Returns:
            Iterable[TaskId]: 子任务ID列表
        """
        pass

    @abstractmethod
    def get_parents_id(self, task: TaskLike, strict: bool = True) -> Iterable[TaskId]:
        """
        获取任务的所有父任务ID

        Args:
            task (TaskLike): 任务对象
            strict (bool, optional): 是否严格模式. Defaults to True.

        Returns:
            Iterable[TaskId]: 父任务ID列表
        """
        pass

    @abstractmethod
    def has_parent_child_relationship(
        self, a: TaskLike, b: TaskLike, strict: bool = True
    ) -> bool:
        """
        判断任务a是否是任务b的父任务

        Args:
            a (TaskLike): 任务A
            b (TaskLike): 任务B
            strict (bool): 是否严格判断. 如果严格判断, 则判断是否存在一条有向边从a指向b. 否则判断是否存在一条从a到b的路径.

        Returns:
            bool: 是否是子任务
        """
        pass

    @abstractmethod
    def update_task_status(self, task: TaskId, status: TaskStatus):
        """
        更新任务状态
        Args:
            task (TaskId): 任务ID
            status (TaskStatus): 任务状态
        """
        pass


class TaskGraph(AbstractTaskManager):
    """
    以任务id作为节点id组成的任务树,内部数据结构是一个有向无环图
    """

    __KEY_TASK__ = "task"

    def __init__(self, **attr):
        self.__graph__ = DiGraph()
        self.__id_task_map__ = defaultdict()

    def parse_task_id(self, task: TaskLike) -> TaskId:
        if isinstance(task, (str, int)):
            return task
        if isinstance(task, ITask):
            return task.get_task_id()
        raise ValueError(f"Unknown task type {type(task)}")

    def task_size(self) -> int:
        """
        获取任务数量
        """
        return self.__graph__.number_of_nodes()

    def tasks(self, only_id: bool = False) -> Iterator[ITask]:
        for node_id in self.__graph__.nodes:
            yield node_id if only_id else self.get_task(node_id)

    def get_task(self, task_id: TaskId) -> Optional[ITask]:
        task = self.__id_task_map__.get(task_id)
        if task:
            return task
        node = self.__graph__.nodes.get(task_id)
        return node.get(self.__KEY_TASK__) if node else None

    def find_by(self):
        raise NotImplementedError("TaskGraph.find_by not implemented")

    def add_task(self, task: TaskLike, parent: Optional[TaskLike] = None, **attr):
        task_id = self.parse_task_id(task)
        self.__id_task_map__[task_id] = task
        self.__graph__.add_node(task_id, **attr)
        if parent:
            parent_id = self.parse_task_id(parent)
            self.add_task_if_absent(parent)
            assert (
                parent_id != task_id
            ), f"parent task {parent_id} and task itself {task_id} should be different"
            self.__graph__.add_edge(parent_id, task_id)

    def add_task_if_absent(
        self, task: TaskLike, parent: Optional[TaskLike] = None, **attr
    ) -> bool:
        task_id = self.parse_task_id(task)
        has = self.__graph__.has_node(task_id)
        if not has:
            self.add_task(task, parent=parent, **attr)
        return has

    def remove_task(self, task: TaskLike, cascade: bool = False, **kwargs):
        task_id = self.parse_task_id(task)
        if cascade:
            for child_id in list(self.get_children_id(task, strict=True)):
                self.remove_task(child_id, cascade=cascade, **kwargs)
        self.__graph__.remove_node(task_id)

    def remove_task_if_present(
        self, task: TaskLike, cascade: bool = False, **kwargs
    ) -> bool:
        task_id = self.parse_task_id(task)
        has = self.__graph__.has_node(task_id)
        if has:
            self.remove_task(task, cascade=cascade, **kwargs)
        return has

    def get_children_id(self, task: TaskLike, strict: bool = True) -> Iterator[TaskId]:
        if strict:
            return self.__graph__.successors(self.parse_task_id(task))
        raise NotImplementedError()

    def get_parents_id(self, task: TaskLike, strict: bool = True) -> Iterable[TaskId]:
        if strict:
            return self.__graph__.predecessors(self.parse_task_id(task))
        raise NotImplementedError()

    def has_parent_child_relationship(
        self, a: TaskLike, b: TaskLike, strict: bool = True
    ) -> bool:
        a_id = self.parse_task_id(a)
        b_id = self.parse_task_id(b)
        return (
            self.__graph__.has_edge(a_id, b_id)
            if strict
            else has_path(self.__graph__, a_id, b_id)
        )

    def __contains__(self, task: TaskLike) -> bool:
        """
        检查任务是否存在
        """
        return self.__graph__.has_node(self.parse_task_id(task))

    def update_task_status(self, task: TaskId, status: TaskStatus):
        """
        将任务状态设置为指定值
        """
        task_id = self.parse_task_id(task)
        task: ITask = self.get_task(task_id)
        assert task, f"任务 {task_id} 不存在"
        task.update_status(status)

    def is_status_at(self, task, status, infer=True, update=True):
        task_id = self.parse_task_id(task)
        task: ITask = self.get_task(task_id)

        match = task.is_status_at(status)
        if match is True:
            return match
        if not infer:
            return match

        children_ids = list(self.get_children_id(task_id, strict=True))
        all_children_match = (
            False
            if not children_ids
            else all(
                self.is_status_at(child_id, status, infer=infer, update=update)
                for child_id in children_ids
            )
        )
        if all_children_match and update:
            self.update_task_status(task_id, status)
        return all_children_match

    def is_task_finished(self, task: TaskLike, update: bool = True, infer=True):
        return self.is_status_at(task, TaskStatus.FINISHED, infer=infer, update=update)



from pathlib import Path
from typing import List, Union

import i18n

from .locales import *

INNER_LOCALES_DIR = Path(__file__).parent.joinpath("locales")

# 设置默认语言
i18n.set("locale", "zh")
i18n.set("fallback", "en")
i18n.set("filename_format", "{locale}.{format}")
i18n.set("skip_locale_root_data", True)
if INNER_LOCALES_DIR.exists():
    i18n.load_path.append(str(INNER_LOCALES_DIR))


def add_locale_dir(path: Union[str, List[str]]):
    """
    添加语言包目录
    Args:
        path: 语言包目录
    """
    if isinstance(path, (str, Path)):
        paths = [path]
    else:
        paths = path

    for p in paths:
        i18n.load_path.append(p)



import logging
from abc import abstractmethod
from typing import TypeVar

import simpy
from ipa.decorator import deprecated

from logis.biz.sim.const import OperationType
from logis.biz.sim.stock.model import IStock, QuantifiedStock
from logis.data_type.point import Point
from logis.data_type.unitable import unify_quantified_value
from logis.iface import Storable
from logis.math import euclid_distance

from .base import *
from .container import *
from .retrieve import *
from .store import *


class IStorage(Storable):
    """
    存储设备
    """

    store_strategy: Optional[StoreStrategy] = StoreStrategy()
    retrieve_strategy: Optional[RetrieveStrategy] = RetrieveStrategy()

    props: Optional[StorageProperties] = None

    __resource__: Optional[simpy.Resource] = None

    # 表明当前存储空间是否被占用
    __occupied__: Optional[simpy.Resource] = None

    __container__: Optional[QuantifiedContainer] = None

    log: logging.Logger

    @property
    @abstractmethod
    def env(self) -> simpy.Environment:
        pass

    def __init__(
        self, props: StorageProperties, exclusive: bool = True, *args, **kwargs
    ):
        """
        Args:
            exclusive: 是否独占，适用于储位、货架、货架组等
        """
        self.props = props
        # 这里以一个不可能达到的数值表示共享
        self.__occupied__ = simpy.Resource(self.env, 1 if exclusive else 10**12)
        p = self.props
        self.__container__ = QuantifiedContainer(p.capacity, env=self.env)
        self.center_point: Optional[Point] = None
        self.current_jobs: int = 0

    def decrease_jobs(self, *args, **kwargs):
        """
        减少作业数
        """
        self.current_jobs -= 1

    def increase_jobs(self, *args, **kwargs):
        """
        增加作业数
        """
        self.current_jobs += 1

    @deprecated("use agent.distance_to instead")
    def distance_to(self, **kwargs) -> float:
        """
        获取到货物源头的距离
        TODO：其实这里按理说应该包括取货距离和送货距离，且放在这里似乎也不太合适
        """
        from logis.biz.sim.agent import IAgent

        # TODO: 考虑过stock不再继承IAgent，所以这里也要考虑改变
        x: IAgent = kwargs.get("agent", None) or kwargs.get("stock", None)
        return euclid_distance(self.center_point, x.current_location)

    @property
    def level(self):
        return self.__container__.level

    @property
    def capacity(self):
        return self.__container__.capacity

    @property
    def unit(self):
        return self.__container__.unit

    @property
    def free_capacity(self):
        return self.__container__.free_capacity

    def store(self, v: QuantifiedStock, *args, **kwargs):
        """
        存储
        """
        log = self.log
        v = unify_quantified_value(v, self.unit)
        cells = self.store_strategy.find_location(self.props, [v])
        with self.__occupied__.request() as req:
            # yield req
            if self.free_capacity < v.quantity:
                log.warning("%s < %s, skip store", self.free_capacity, v.quantity)
                yield self.env.timeout(0)
                return
            for cell in cells:
                cell.store(v)
            yield self.__container__.put(v)
            t = self.props.store_speed_time
            yield self.env.timeout(0 if not t else t.value)

    def retrieve(self, v: QuantifiedStock, *args, **kwargs):
        """
        取回
        """
        log = self.log
        v = unify_quantified_value(v, self.unit)
        cells = self.retrieve_strategy.find_location(self.props, [v])
        with self.__occupied__.request() as req:
            yield req
            if v.quantity > self.level:
                log.warning("%s < %s, skip retrieve", self.level, v.quantity)
                yield self.env.timeout(0)
                return
            for cell in cells:
                cell.retrieve(v)
            yield self.__container__.get(v)
            t = self.props.retrieve_speed_time or self.props.store_speed_time
            yield self.env.timeout(0 if not t else t.value)


class ICell(IStorage):
    """
    储位
    """

    @property
    @abstractmethod
    def id(self):
        pass

    @property
    @abstractmethod
    def is_still_universal(self) -> bool:
        pass

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.current_quantity: Optional[simpy.Container] = None
        self.target_quantity: Optional[simpy.Container] = None

    @abstractmethod
    def is_able_to(
        self, operation: OperationType, stock: IStock, strict: bool = False, **kwargs
    ) -> bool:
        """
        判断是否可操作
        Args:
            operation: 操作类型
            stock: 货物
            strict: 是否严格判断，即完全能满足所有货物的需求
        Returns:
            是否可操作
        """
        pass


CellClass = TypeVar("CellClass", bound=ICell)


class IRack(IStorage):
    """
    货架
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    @abstractmethod
    def cells(self) -> List[ICell]:
        """
        获取本货架下的所有储位
        """
        pass

    @abstractmethod
    def is_able_to(
        self, operation: OperationType, stock: IStock, strict: bool = False, **kwargs
    ) -> bool:
        """
        判断是否能完成某个操作
        Args:
            operation: 操作类型
            stock: 货物
            strict: 是否严格判断，即完全能满足所有货物的需求
        Returns:
            是否能完成操作
        """


RackClass = TypeVar("RackClass", bound=IRack)


class IRackGroup(IStorage):
    """
    货架组
    """

    @property
    @abstractmethod
    def racks(self) -> List[IRack]:
        """
        获取本货架组下的所有货架
        """
        pass


RackGroupClass = TypeVar("RackGroupClass", bound=IRackGroup)



import asyncio
import logging
import time
from datetime import datetime

import pytest

from logis.util.io_util import AsyncWriteBuffer, SyncWriteBuffer, WriteBufferConfig

logging.getLogger().setLevel(logging.DEBUG)


def test_sync_write_buffer():
    config = WriteBufferConfig(
        batch_size=2,
        flush_interval=1,
        handler=lambda items: logging.debug(f"{datetime.now()}处理 {items}"),
    )
    buffer = SyncWriteBuffer(config)
    buffer.start()

    for i in range(5):
        buffer.put(datetime.now())
        time.sleep(0.5)

    buffer.stop()


@pytest.mark.asyncio
async def test_async_write_buffer():
    config = WriteBufferConfig(
        batch_size=2,
        flush_interval=1,
        handler=lambda items: print(f"处理 {items}"),
    )
    buffer = AsyncWriteBuffer(config)

    asyncio.create_task(buffer.start())

    for i in range(5):
        await buffer.put(datetime.now())
        await asyncio.sleep(0.5)

    await buffer.stop()



from abc import ABC, ABCMeta, abstractmethod
from typing import Generic, TypeVar

T = TypeVar("T")


class IRegistry(ABC, Generic[T]):

    @abstractmethod
    def register(self, *args, **kwargs):
        pass

    @abstractmethod
    def unregister(self, *args, **kwargs):
        pass

    @abstractmethod
    def get[T](self, *args, **kwargs) -> T:
        pass



import time
from ast import Dict
from pathlib import Path
from typing import Optional

from sqlalchemy import create_engine
from sqlalchemy.types import JSON
from sqlmodel import Field, Session, SQLModel, delete, func, select, update

from logis.data_type import DEFAULT_PYDANTIC_MODEL_CONFIG

DATA_DIR = Path(__file__).parent.parent / "data"
assert DATA_DIR.exists()


class Event(SQLModel, table=True):
    __tablename__ = "event"
    id: Optional[int] = Field(
        default=None, primary_key=True, sa_column_kwargs={"autoincrement": True}
    )
    time: float
    type: str
    data: str

    model_config = DEFAULT_PYDANTIC_MODEL_CONFIG

from uuid import uuid4

db_file = f"sqlite:///{DATA_DIR}/test.db"
engine = create_engine(db_file, echo=True)
SQLModel.metadata.create_all(engine)


def test_insert():

    s = Session(bind=engine)

    events = []
    size = 10
    for i in range(size):
        event = Event(time=i, type="test", data="test")
        events.append(event)

    start = time.time()
    s.bulk_save_objects(events)
    s.commit()
    end = time.time()
    print(f"insert {size} events cost {end-start} seconds")

    s.close()


def test_select():
    s = Session(bind=engine, close_resets_only=True)
    start = time.time()
    size = s.exec(select(func.count()).where(Event.id < 100000)).first()
    dt1 = time.time() - start
    # close_resets_only=True意味着close未必立即关闭还是有被复用的可能
    s.close()
    print(f"select {size} events cost {dt1} seconds")

    start = time.time()
    tmps = s.query(Event).filter(Event.id < 100000).all()
    dt2 = time.time() - start
    print(f"query {len(tmps)} events cost {dt2} seconds")

    print(f"{dt1} < {dt2}")



import pytest

from logis.util.lambda_util import invoke


class A:
    @property
    def id(self) -> int:
        pass


class B(A):
    def __init__(self):
        self.id = 10


def test_id():
    with pytest.raises(AttributeError):
        b = B()
        assert b.id == 10


def test_mro():
    class A:
        def __init__(self, **kwargs):
            print("->A")
            super().__init__(**kwargs)
            print("->A end")

        def method(self):
            return "A"

    class B(A):
        def __init__(self, **kwargs):
            print("->B")
            super().__init__(**kwargs)
            print("->B end")

    class C(A):
        def __init__(self, **kwargs):
            print("->C")
            super().__init__(**kwargs)
            print("->C end")

        def method(self):
            return "C"

    class D(B, C):
        def __init__(self, **kwargs):
            print("->D")
            super().__init__(**kwargs)
            print("->D end")

    assert D().method() == "C"
    assert D.mro() == [D, B, C, A, object]


def test_atexit():
    from atexit import register

    async def hello(name: str):
        return print("hello " + name)

    def handler():
        print("atexit111")

    register(handler)

    class MyClass:
        def __init__(self):
            register(lambda: invoke(hello, "andy"))

    MyClass()



from typing import Any, Literal, TypeVar

from logis.data_type.unitable import NumberUnit


def split_integer(
    total: int,
    n: int,
    mode: Literal["limit_parts", "limit_per_part"] = "limit_parts",
    order: Literal["asc", "desc"] = "desc",
):
    """
    将整数尽可能平均分成parts份,每份都需要是整数

    Args:
        total: 总整数
        n: 需要分成的份数或每份的值
        mode: 模式，limit_parts表示将total尽可能平均分成n份，limit_per_part表示每份都需要是n
        order: 排序方式
    """
    n = int(n)
    assert n > 0, "份数必须大于0"

    arrs = []
    if mode == "limit_parts":
        # 基础值和余数
        base = total // n  # 每份的基础值
        remainder = total % base  # 需要多分配1的份数
        arrs = [
            (remainder, base + 1),
            (n - remainder, base),
        ]
    elif mode == "limit_per_part":
        # 基础值和余数
        base = n  # 每份的基础值
        remainder = total % base
        arrs = [
            (total // base, base),
        ]
        if remainder:
            arrs.append((1, remainder))
    else:
        raise ValueError("mode must be limit_parts or limit_per_part")

    if order == "asc":
        arrs.reverse()
    for times, value in arrs:
        for _ in range(int(times)):
            yield value


V = TypeVar("V")


def get_numeric_value(some: V):
    """
    获取输入对象的数值
    """
    if isinstance(some, NumberUnit):
        return some.quantity
    return some



from abc import ABCMeta, abstractmethod
from collections import Counter, defaultdict
from enum import Enum
from numbers import Number
from typing import Dict, Iterable, Optional, Union

from pydantic import BaseModel, Field

from logis.data_type import NumberType

TaskId = Union[str, int]
TaskPriority = Union[str, int]


class TaskStatus(Enum):
    """
    任务状态
    """

    NOT_STARTED = "not_started"
    STARTED = "started"
    PAUSED = "paused"
    FINISHED = "finished"
    CANCELLED = "cancelled"
    FAILED = "failed"


class ITask(metaclass=ABCMeta):
    """
    任务基础抽象类
    """

    @abstractmethod
    def get_task_id(self) -> TaskId:
        pass

    @abstractmethod
    def get_priority(self) -> TaskPriority:
        """
        任务优先级
        """
        pass

    @abstractmethod
    def update_status(self, status: TaskStatus):
        """
        更新任务状态
        """
        pass

    @abstractmethod
    def is_status_at(self, status: TaskStatus) -> bool:
        """
        检查任务状态是否为指定值
        """
        pass

    @property
    def finished(self) -> bool:
        return self.is_status_at(TaskStatus.FINISHED)


from logis.data_type import NumberUnit


class QuantifiedTask(NumberUnit, ITask):
    """
    可量化的任务
    """

    progress: Union[Number] = 0

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.pipeline_progress: Dict[str, Number] = Counter()

    def update_progress(
        self,
        progress: Union[NumberUnit, Number],
        delta: bool = False,
        pipeline_id: Optional[str] = None,
    ):
        """
        更新进度

        Args:

        """
        if isinstance(progress, NumberUnit):
            assert (
                progress.unit == self.unit
            ), f"进度单位{progress.unit}与任务单位{self.unit}不一致，请在调用处处理"
            progress = progress.quantity

        if pipeline_id is not None:
            if delta:
                self.pipeline_progress[pipeline_id] += progress
            else:
                self.pipeline_progress[pipeline_id] = progress
        else:
            if delta:
                self.progress += progress
            else:
                self.progress = progress


TaskLike = Union[ITask, TaskId]


class Task(BaseModel, ITask):
    """
    任务基础结构
    """

    id: Optional[TaskId] = None
    parent_id: Optional[TaskId] = None
    children_ids: Iterable[TaskId] = Field(default_factory=list)

    name: Optional[str] = None
    type: Optional[str] = None
    priority: Optional[TaskPriority] = None

    repeat: int = 1
    progress: Union[float, int] = 0
    tags: Iterable[str] = Field(default_factory=set)

    created_at: Optional[int] = None
    updated_at: Optional[int] = None
    started_at: Optional[int] = None
    finished_at: Optional[int] = None
    cancelled_at: Optional[int] = None

    status: Optional[TaskStatus] = None
    stage: Optional[str] = None

    remark: Optional[str] = None

    def get_task_id(self) -> TaskId:
        return self.id

    def get_priority(self) -> TaskPriority:
        return self.priority

    def update_status(self, status: TaskStatus):
        self.status = status

    def is_status_at(self, status: TaskStatus) -> bool:
        return self.status == status



from logis.util import cast_if_not_none, round_if_not_none, set_attr_if, set_attr_if_has
from logis.util.dict_util import remove_key_not_in
from logis.util.num_util import split_integer


def test_round_if_not_none():
    assert round(3.745, 2) == 3.75
    assert round(3.74512, 2) == 3.75
    assert round(1216.745, 2) == 1216.74
    assert round(1216.74512, 2) == 1216.75

    assert round_if_not_none(1216.745, 2) == 1216.75
    assert round_if_not_none(1216.74512, 2) == 1216.75
    assert round_if_not_none(3.745, 2) == 3.75
    assert round_if_not_none(3.74512, 2) == 3.75

    assert round_if_not_none(3.74512, None) == 3.74512
    assert round_if_not_none(None, 2) is None
    assert round_if_not_none(None, None) is None


def test_cast_if_not_none():
    assert cast_if_not_none(3.14159, int) == 3
    assert cast_if_not_none(3.14159, float) == 3.14159
    assert cast_if_not_none(None, int) is None
    assert cast_if_not_none(None, float) is None


def test_split_integer():
    assert list(split_integer(10, 3, order="asc", mode="limit_parts")) == [3, 3, 4]
    assert list(split_integer(10, 3, order="desc", mode="limit_parts")) == [4, 3, 3]

    assert list(split_integer(10, 3, order="desc", mode="limit_per_part")) == [
        3,
        3,
        3,
        1,
    ]
    assert list(split_integer(10, 3, order="asc", mode="limit_per_part")) == [
        1,
        3,
        3,
        3,
    ]


def test_remove_key_not_in():
    dc = dict(a=1, b=2, c=3, d=4)
    remove_key_not_in(dc, dict(d=5, e=6).keys())
    assert list(dc.keys()) == ["d"]


class Animal:
    def __init__(self):
        self.name = None
        self.age = None


def test_set_attr_if():

    a = Animal()
    set_attr_if_has(
        a,
        dict(name="dog", age=10, gender=1),
    )

    assert a.name == "dog"
    assert a.age == 10
    assert getattr(a, "gender", None) is None

    set_attr_if(a, dict(gender=1), lambda *args: True)

    assert hasattr(a, "gender") and getattr(a, "gender") == 1



from abc import ABCMeta, abstractmethod
from collections import defaultdict
from fractions import Fraction
from functools import reduce
from typing import Any, Dict, List, Optional, Self, Tuple, Union

from pydantic import BaseModel

from logis.data_type import DEFAULT_PYDANTIC_MODEL_CONFIG as MODEL_CONFIG
from logis.data_type import NumberType, Unit

from .point import *


class NumberUnit(metaclass=ABCMeta):
    """
    具有数量和单位的数值类型
    """

    quantity: NumberType = 0
    unit: Optional[Unit] = None

    # 自定义倍率转换器
    _unit_config_: Optional["UnitConfig"] = None

    def __init__(self, **kwargs):
        self.quantity = kwargs.get("quantity", 0)
        self.unit = kwargs.get("unit", None)

    @property
    def kind(self):
        """
        默认情况下返回类名,实际使用时可以重写
        """
        return self.__class__.__name__

    @abstractmethod
    def model_dump(self, *args, **kwargs) -> Dict[str, Any]:
        pass

    @classmethod
    @abstractmethod
    def model_validate(cls, data: Dict[str, Any], **kwargs) -> Self:
        pass

    def __auto_validate__(self, other: "NumberUnit"):
        if self._unit_config_ is None:
            assert (
                self.unit == other.unit
            ), f"unit must be the same, but {self.unit} != {other.unit}"
        else:
            other = unify_quantified_value(
                other, self.unit, unit_config=self._unit_config_
            )
        return other

    def __sub__(self, other: "NumberUnit") -> Self:
        other = self.__auto_validate__(other)
        dc = self.model_dump()
        dc["quantity"] = self.quantity - other.quantity
        return type(self).model_validate(dc)

    def __add__(self, other: "NumberUnit") -> Self:
        other = self.__auto_validate__(other)
        dc = self.model_dump()
        dc["quantity"] += other.quantity
        return type(self)(**dc)

    def __truediv__(self, other: "NumberUnit") -> Self:
        other = self.__auto_validate__(other)
        # TODO: 处理精度问题
        return self.quantity / other.quantity

    def __mul__(self, times: Union[int, float]) -> Self:
        data = self.model_dump()
        data["quantity"] *= times
        return type(self).model_validate(data)

    def __gt__(self, other: "NumberUnit"):
        other = self.__auto_validate__(other)
        return self.quantity > other.quantity

    def __lt__(self, other: "NumberUnit"):
        other = self.__auto_validate__(other)
        return self.quantity < other.quantity

    def __eq__(self, other: "NumberUnit"):
        if not other:
            return False
        other = self.__auto_validate__(other)
        return self.quantity == other.quantity

    def __ge__(self, other: "NumberUnit"):
        other = self.__auto_validate__(other)
        return self.quantity >= other.quantity

    def __le__(self, other: "NumberUnit"):
        other = self.__auto_validate__(other)
        return self.quantity <= other.quantity

    @property
    def value(self):
        return self.quantity

    @value.setter
    def value(self, new_value):
        self.quantity = new_value

    @classmethod
    def parse_tuple(cls, input: Tuple[NumberType, Unit]) -> Self:
        """
        解析形如[1,"个"]的数据
        """
        return cls(quantity=input[0], unit=input[1])

    @classmethod
    def parse_str(cls, input: str, delimiter="|", number_type=float) -> Optional[Self]:
        """
        解析形如1|m/s的数据
        """
        if not input:
            return None
        tmps = input.strip().split(delimiter)
        assert len(tmps) == 2, "unexpected format: " + input
        return cls(quantity=number_type(tmps[0]), unit=tmps[1])

    @classmethod
    def of(cls, num: NumberType, unit: Optional[str] = None) -> Self:
        return cls(quantity=num, unit=unit)

    def increase(self, num: NumberType):
        self.quantity += num

    def decrease(self, num: NumberType):
        self.quantity -= num


class QuantifiedValue(BaseModel, NumberUnit):
    """
    量化值，相比于接口NumberUnit，适合作为数据结构使用
    """
    model_config = MODEL_CONFIG

    name: Optional[str] = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class Capacity(QuantifiedValue):
    pass


class Length(QuantifiedValue):
    pass


class Speed(QuantifiedValue):
    pass


SpeedVector = Tuple[Optional[Speed], Optional[Speed], Optional[Speed]]


class ThreeDimensionalVelocity(BaseModel):
    """
    使用x,y,z表示的三维速度
    """

    model_config = MODEL_CONFIG

    x: Optional[Speed] = None
    y: Optional[Speed] = None
    z: Optional[Speed] = None

    @classmethod
    def of(
        cls,
        x: Optional[NumberType] = None,
        y: Optional[NumberType] = None,
        z: Optional[NumberType] = None,
        unit: Optional[str] = None,
    ):
        return cls(
            x=None if x is None else Speed.of(x, unit),
            y=None if y is None else Speed.of(y, unit),
            z=None if z is None else Speed.of(z, unit),
        )


class Time(QuantifiedValue):
    """
    时间
    """

    pass


def get_time(l: Length, v: Speed):
    """
    距离/速度=时间
    TODO：处理单位
    """
    assert l.value is not None, "length is not specified"
    assert v.value is not None, "speed is not specified"
    assert v.value != 0, "speed cannot be zero"
    return abs(l.value / v.value)


def get_time_3d(delta_distance: Point, v: ThreeDimensionalVelocity):
    """
    分别计算各维度的时间，取最大值
    """
    lz, ly, lx, lunit = (
        delta_distance.z,
        delta_distance.y,
        delta_distance.x,
        delta_distance.unit,
    )
    return max(
        0 if not lx else get_time(Length(quantity=lx, unit=lunit), v.x),
        0 if not ly else get_time(Length(quantity=ly, unit=lunit), v.y),
        0 if not lz else get_time(Length(quantity=lz, unit=lunit), v.z),
    )


DEFAULT_UNIT_CONFIG = dict()

from pint import UnitRegistry, set_application_registry

# 1. 初始化单位注册表
ureg = UnitRegistry()
# 设置为全局注册表（避免重复注册）
set_application_registry(ureg)

Quantity = ureg.Quantity

# 2. 将中文与内置单位关联
# 长度单位
ureg.define("米 = meter")
ureg.define("分米 = decimeter")
ureg.define("厘米 = centimeter")
ureg.define("毫米 = millimeter")
ureg.define("纳米 = nanometer")
ureg.define("千米 = kilometer")
ureg.define("公里 = kilometer")

# 重量单位
ureg.define("千克 = kilogram")
ureg.define("克 = gram")
ureg.define("吨 = tonne")

# 时间单位
ureg.define("年 = year")
ureg.define("月 = month")
ureg.define("周 = week")
ureg.define("星期 = week")
ureg.define("天 = day")
ureg.define("小时 = hour")
ureg.define("分钟 = minute")
ureg.define("秒 = second")

# 温度单位
ureg.define("摄氏度 = degree_Celsius")
ureg.define("华氏度 = degree_Fahrenheit")


def get_unit_ratio(src: Unit, dst: Unit, src_quantity: NumberType = 1) -> float:
    """
    获取 src 到 dst 的倍率
    """
    src, dst = src.replace("每", "/"), dst.replace("每", "/")
    return ureg.Quantity(src_quantity, src).to(dst).magnitude


class UnitConfig(dict):
    """
    有关单位的配置项
    """

    def get_ratio(self, src: Unit, dst: Unit, src_quantity: NumberType = 1) -> Fraction:
        """
        获取 src 到 dst 的倍率
        """
        if src == dst:
            return Fraction(src_quantity, 1)
        if src in self and dst in self:
            return Fraction(self[dst] * src_quantity, self[src])
        try:
            return Fraction(get_unit_ratio(src, dst, src_quantity=src_quantity))
        except:
            pass
        raise ValueError(f"unknown unit: {src} or {dst}")

    def get_float_ratio(self, src: Unit, dst: Unit, *args, **kwargs) -> float:
        """
        获取 src 到 dst 的倍率
        """
        return float(self.get_ratio(src, dst, *args, **kwargs))

    def get_int_ratio(self, src: Unit, dst: Unit, *args, **kwargs) -> int:
        """
        获取 src 到 dst 的倍率
        """
        return int(self.get_ratio(src, dst, *args, **kwargs))

    def alias(self, unit: Unit, *aliases: Unit):
        """
        为单位添加别名
        """
        for alias in aliases:
            self[alias] = self.get(unit)
        return self

    def __or__(self, value: Dict):

        obj = dict(value)
        obj.update(self)

        return UnitConfig(obj)


class UnitConfigBuilder:
    """
    单位配置构建器
    """

    def __init__(self, base_unit: Unit):
        self.__unit_config__ = UnitConfig([[base_unit, 1]])

    @staticmethod
    def with_base(unit: Unit):
        return UnitConfigBuilder(unit)

    def add(self, unit: Unit, multiplier_to_base: NumberType):
        """
        Args:
            unit: 单位
            multiplier_to_base: 转换到1base单位所需要的倍率，例如：base为1m时，cm到1m需要100倍
        """
        self.__unit_config__[unit] = multiplier_to_base
        return self

    def alias(self, unit: Unit, *aliases: Unit):
        """
        为单位添加别名
        """
        self.__unit_config__.alias(unit=unit, *aliases)
        return self

    def build(self) -> UnitConfig:
        return self.__unit_config__


TIME_UNIT_CONFIG = UnitConfig(
    天=1, 小时=24, 分钟=24 * 60, 秒=24 * 60 * 60, 毫秒=24 * 60 * 60 * 10**3
)
VOLUME_UNIT_CONFIG = UnitConfig(kl=Fraction(1, 10**3), l=1, dl=10, cl=100, ml=10**3)
LENGTH_UNIT_CONFIG = UnitConfig(
    km=Fraction(1, 10**3), m=1, dm=10, cm=100, mm=10**3, nm=10**9
)

DEFAULT_UNIT_CONFIG = TIME_UNIT_CONFIG | VOLUME_UNIT_CONFIG | LENGTH_UNIT_CONFIG


def unify_quantified_value(
    self: QuantifiedValue,
    target_unit: Optional[Unit] = None,
    unit_config: UnitConfig = DEFAULT_UNIT_CONFIG,
    modify_ref: bool = False,
):
    """
    单位转换
    """
    same_unit = self.unit == target_unit
    if unit_config is None:
        assert same_unit, "unit must be the same if no radio_computer given"
    else:
        ratio = 1 if same_unit else unit_config.get_ratio(self.unit, target_unit)
        if modify_ref:
            self.quantity = self.quantity * ratio
            self.unit = target_unit
            return self
        # 上面这种方式能尽可能避免属性丢失，但修改的是原始对象
        dc = {
            **self.model_dump(),
            **dict(quantity=self.quantity * ratio, unit=target_unit),
        }
        return type(self)(**dc)


def merge_quantified_value(
    items: List[QuantifiedValue],
    unit_config: Optional[UnitConfig] = DEFAULT_UNIT_CONFIG,
    target_unit: Optional[Unit] = None,
):
    """
    将所有的物理量合并，这里是假定所有的物理量都是同类，例如：全是长度，而不是既有长度又有速度

    Args:
        items: 待合并项
        ratio_computer: 倍率计算器
        target_unit: 目标单位，如果未指定，默认使用数组的第一个条目的单位
    """
    assert items, "item length need to be at least 1"
    target_unit = target_unit or items[0].unit
    if len(items) != 2:
        return reduce(
            lambda a, b: merge_quantified_value(
                [a, b], target_unit=target_unit, unit_config=unit_config
            ),
            items,
        )
    self, other = unify_quantified_value(
        items[0], target_unit=target_unit, unit_config=unit_config
    ), unify_quantified_value(
        items[1], target_unit=target_unit, unit_config=unit_config
    )
    dc = {
        **self.model_dump(),
        **dict(quantity=self.quantity + other.quantity, unit=target_unit),
    }
    return type(self)(**dc)


def group_merge_quantified_value(items: List[QuantifiedValue]) -> List[QuantifiedValue]:
    """
    按照单位分组，并合并同类单位物理量
    """
    result = []
    unit_items_map = defaultdict(list)
    for item in items:
        unit_items_map[item.unit].append(item)
    for _, tmps in unit_items_map.items():
        result.append(merge_quantified_value(tmps))
    return result



__doc__ = "寻路算法模块"


from pathfinding.finder.finder import Finder

from logis.math import *

from ._a_star import AStarPathFinding
from .base import *
from .model import *

# from ._dijkstra import DijkstraPathFinding


if __name__ == "__main__":
    pass



from typing import Optional

from pydantic import AliasChoices, BaseModel, Field, field_validator

from logis.data_type.base import DEFAULT_PYDANTIC_MODEL_CONFIG
from logis.data_type.point import Point
from logis.data_type.unitable import Length, QuantifiedValue


class SpatialProps(BaseModel):
    """
    空间属性
    """

    model_config = DEFAULT_PYDANTIC_MODEL_CONFIG

    width: Optional[Length] = Field(
        default=None,
        validation_alias=AliasChoices("单元格宽度", "width"),
    )
    height: Optional[Length] = Field(
        default=None, validation_alias=AliasChoices("层高", "height")
    )
    depth: Optional[Length] = Field(
        validation_alias=AliasChoices("货架深度", "托盘货架深度"), default=None
    )
    rotation: Optional[QuantifiedValue] = Field(
        validation_alias=AliasChoices("旋转", "rotate"), default=None
    )
    center_point: Optional[Point] = None

    @field_validator("center_point", mode="before")
    def center_point_validator_before(cls, v: Optional[Point]):
        if v is None:
            return None
        if isinstance(v, Point):
            return v
        if isinstance(v, dict):
            return Point.model_validate(v)
        elif isinstance(v, (tuple, list)):
            return Point.from_tuple(v)
        raise ValueError("only dict, tuple, list type is supported")



from typing import TYPE_CHECKING

from .base import (
    PathFindingAlgorithm,
)
from .model import PathFindingInput, PathFindingOutput

if TYPE_CHECKING:
    pass


class CbsPathFinding(PathFindingAlgorithm):
    """
    CBS寻路算法实现
    """

    type = "cbs"

    def find_path(self, input: PathFindingInput, **kwargs) -> PathFindingOutput:
        """
        使用CBS算法寻找从start到end的路径
        """
        raise NotImplementedError("CBS算法未实现")
        from mapf.centralized.cbs import CBS, Environment
        from mapf.model import AgentInfo, MapfInput, MapfOutput, MapInfo

        env = Environment.from_mapf_input()
        MapfInput(
            map=MapInfo(
                dimensions=input.graph,
                obstacles=list(map(lambda p: p.to_tuple(), input.excluded_vertices)),
            ),
            agents=[
                AgentInfo(
                    name=0,
                    start=input.start.to_tuple(),
                    goal=input.end.to_tuple(),
                )
            ],
        )

        cbs = CBS(env)
        out = cbs.search()



from abc import ABC, abstractmethod
from typing import Any, Dict, Literal, Optional

from networkx import DiGraph

from logis.data_type import UnitConfig


class IDataReport(ABC):
    pass


class IJsonParser(ABC):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.logic_graph: Optional[DiGraph] = None
        self._produce_recipe_graph: Optional[DiGraph] = None
        self.object_map: Optional[Dict[str, Any]] = None


class IResultGenerator(ABC):
    pass


class IExcelParser(ABC):
    pass


class IOrderManager(ABC):
    pass


class IUnitManager(ABC):

    @abstractmethod
    def find_unit_level(
        self, stock_name: str, unit_name: str
    ) -> Optional[Literal["一级单位", "二级单位", "三级单位"]]:
        pass

    @abstractmethod
    def get_unit_config(self, stock_name: str) -> UnitConfig:
        pass


class IResourceManager(ABC):
    pass


class IStorageManager(ABC):
    pass



from datetime import datetime
from typing import TYPE_CHECKING, Dict, List, Optional

from ipa.decorator import deprecated
from pydantic import BaseModel
from sqlmodel import Field, SQLModel

from logis.biz.sim.produce import RecipeItemVo
from logis.data_type import DEFAULT_PYDANTIC_MODEL_CONFIG
from logis.event import EventBaseModel
from logis.metric import MetricModel
from logis.metric.influxdb import Point


class BaseEventMetric(EventBaseModel):
    """
    base event metric.
    """

    # 数据所属用户
    user_id: Optional[str] = None

    # 每次仿真有不同的identifier，便于区分
    sim_id: str

    sim_time: Optional[float] = None

    event_time: Optional[float] = Field(exclude=True, default=None)


@deprecated("useless for now")
class SimEventMetric(BaseEventMetric[str], SQLModel, table=False):
    """

    simualtion system metric event.

    """

    __tablename__ = "metric"

    # 下面是一些常规字段
    created_at: datetime = datetime.now()

    updated_at: Optional[datetime] = None

    id: Optional[int] = Field(
        primary_key=True, default=None, sa_column_kwargs={"autoincrement": True}
    )


class PickingStationMetric(BaseEventMetric, MetricModel):
    """
    picking station metric event.
    """

    sim_id: Optional[str] = None

    picking_station_name: Optional[str] = None

    picking_station_id: Optional[str] = None

    working_mode: Optional[str] = None

    @classmethod
    def get_measurement(
        cls, prefix: Optional[str] = None, suffix: Optional[str] = None, sep: str = "/"
    ):
        return sep.join(filter(lambda x: x, [prefix, "picking_station_metric", suffix]))

    def to_influxdb_point(self, measurement: Optional[str] = None):
        p = Point(measurement or self.get_measurement())
        p.tag("event_type", self.event_type)
        p.tag("sim_id", self.sim_id)
        p.tag("picking_station_id", self.picking_station_id)
        p.field("picking_station_name", self.picking_station_name)
        p.field("working_mode", self.working_mode)
        return p


class TaskEventMetric(BaseEventMetric[str], MetricModel):
    """
    task metric event.
    """

    sim_id: Optional[str] = None

    order_id: Optional[str] = None

    task_id: str

    sim_time: Optional[float] = None

    success: Optional[bool] = None

    @classmethod
    def get_measurement(
        cls, prefix: Optional[str] = None, suffix: Optional[str] = None, sep: str = "/"
    ):
        return sep.join(filter(lambda x: x, [prefix, "task_metric", suffix]))

    def to_influxdb_point(self, measurement: Optional[str] = None):
        p = Point(measurement or self.get_measurement())
        p.tag("event_type", self.event_type)
        p.tag("sim_id", self.sim_id)
        p.tag("order_id", self.order_id)
        p.tag("task_id", self.task_id)
        p.field("sim_time", self.sim_time)
        p.field("success", self.success)
        return p


class ReplenishEventMetric(TaskEventMetric):
    """
    replenish event metric.
    """

    destination: Optional[str] = None

    stock_name: Optional[str] = None

    quantity: Optional[int] = None

    unit: Optional[str] = None

    @classmethod
    def get_measurement(
        cls, prefix: Optional[str] = None, suffix: Optional[str] = None, sep: str = "/"
    ):
        return sep.join(filter(lambda x: x, [prefix, "replenish_metric", suffix]))

    def to_influxdb_point(self, measurement: Optional[str] = None):
        p = super().to_influxdb_point(measurement)
        p.tag("destination", self.destination)
        p.tag("stock_name", self.stock_name)
        p.field("quantity", self.quantity)
        p.field("unit", self.unit)
        return p


class StorageLocationMetric(BaseModel, MetricModel):
    """
    storage location metric event.
    """

    sim_id: Optional[str] = None

    storage_location_id: str

    delta_quantity: Optional[float] = None

    stock_name: Optional[str] = None

    stock_unit: Optional[str] = None

    @classmethod
    def get_measurement(
        cls, prefix: Optional[str] = None, suffix: Optional[str] = None, sep: str = "/"
    ):
        return sep.join(
            filter(lambda x: x, [prefix, "storage_location_metric", suffix])
        )

    def to_influxdb_point(self, measurement: Optional[str] = None):
        p = Point(measurement or self.get_measurement())
        p.tag("sim_id", self.sim_id)
        p.tag("storage_location_id", self.storage_location_id)
        p.field("delta_quantity", self.delta_quantity)
        p.field("stock_name", self.stock_name)
        p.field("stock_unit", self.stock_unit)
        return p


class StorageMetric(BaseModel, MetricModel):
    """
    storage metric event.
    """

    sim_id: Optional[str] = None

    storage_id: str

    storage_type: Optional[str] = None

    capacity: Optional[float] = None

    capacity_occupied: Optional[float] = None

    stock_name: Optional[str] = None

    stock_unit: Optional[str] = None

    @classmethod
    def get_measurement(
        cls, prefix: Optional[str] = None, suffix: Optional[str] = None, sep: str = "/"
    ):
        return sep.join(filter(lambda x: x, [prefix, "storage_metric", suffix]))

    def to_influxdb_point(self, measurement: Optional[str] = None):
        p = Point(measurement or self.get_measurement())
        p.tag("sim_id", self.sim_id)
        p.tag("storage_id", self.storage_id)
        p.tag("storage_type", self.storage_type)
        p.field("capacity", self.capacity)
        p.field("capacity_occupied", self.capacity_occupied)
        p.field("stock_name", self.stock_name)
        p.field("stock_unit", self.stock_unit)
        return p


class AgentEventMetric(BaseEventMetric[str], MetricModel):
    """
    agent metric event.
    """

    sim_id: Optional[str] = None

    agent_id: int

    agent_type: Optional[str] = None

    order_id: Optional[str] = None
    task_id: Optional[str] = None

    sim_time: Optional[float] = None

    @classmethod
    def get_measurement(
        cls, prefix: Optional[str] = None, suffix: Optional[str] = None, sep: str = "/"
    ):
        return sep.join(filter(lambda x: x, [prefix, "agent_metric", suffix]))

    def to_influxdb_point(self, measurement: Optional[str] = None):
        p = Point(measurement or self.get_measurement())
        p.tag("sim_id", self.sim_id)
        p.tag("event_type", self.event_type)
        p.tag("agent_id", self.agent_id)
        p.tag("agent_type", self.agent_type)
        p.field("sim_time", self.sim_time)
        p.tag("order_id", self.order_id).tag("task_id", self.task_id)
        return p


class SimMetric(BaseModel, MetricModel):
    """
    simulation metric event.
    """

    sim_id: str
    model_name: Optional[str] = None
    sim_abs_time: Optional[float] = None
    event_type: str

    @classmethod
    def get_measurement(
        cls, prefix: Optional[str] = None, suffix: Optional[str] = None, sep: str = "/"
    ):
        return sep.join(filter(lambda x: x, [prefix, "sim_metric", suffix]))

    def to_influxdb_point(self, measurement: Optional[str] = None):
        p = Point(measurement or self.get_measurement())
        p.tag("sim_id", self.sim_id)
        p.tag("event_type", self.event_type)
        p.field("model_name", self.model_name)
        p.field("sim_abs_time", self.sim_abs_time)
        return p


class SimQuery(BaseModel):
    """
    模拟查询条件
    """

    sim_id: str


class StorageDeviceQuery(SimQuery):
    """
    存储设备查询条件
    """

    sim_id: str
    rack_ids: Optional[List[str]] = None
    cell_ids: Optional[List[str]] = None


class TransportDeviceQuery(SimQuery):
    """
    运输设备查询条件
    """

    sim_id: str


class InventoryChange(BaseModel):
    """
    库存变化
    """

    stock_name: str
    stock_unit: str
    delta_quantity: float


class ContainerWatchInfo(BaseModel):
    """
    容器监控信息
    """

    parent_id: Optional[str] = None
    id: str
    capacity: Optional[float] = None
    init_level: Optional[float] = None
    level: Optional[float] = None


class AgentWatchInfo(BaseModel):
    """
    智能体工作信息
    """

    agent_id: int
    agent_type: Optional[str] = None
    order_id: Optional[str] = None
    task_id: Optional[str] = None
    duration: Optional[float] = None
    acquired_at: Optional[float] = None
    released_at: Optional[float] = None

    model_config = DEFAULT_PYDANTIC_MODEL_CONFIG


class TaskWatchInfo(BaseModel):
    """
    任务监控信息
    """

    order_id: Optional[str] = None
    task_id: str
    duration: Optional[float] = None
    # None: 未完成,False: 延迟完成,True: 按时完成
    on_time: Optional[bool] = None

    started_at: Optional[float] = None
    finished_at: Optional[float] = None
    success: Optional[bool] = None

    model_config = DEFAULT_PYDANTIC_MODEL_CONFIG


class ReplenishTaskWatchInfo(TaskWatchInfo):
    """
    补货任务监控信息
    """

    destination: Optional[str] = None
    stock_name: str
    quantity: float
    unit: Optional[str] = None

    model_config = DEFAULT_PYDANTIC_MODEL_CONFIG


class OrderWatchInfo(BaseModel):
    """
    订单监控信息
    """

    order_id: str
    order_type: Optional[str] = None
    duration: Optional[float] = None
    # None: 未完成,False: 延迟完成,True: 按时完成
    on_time: Optional[bool] = None

    started_at: Optional[float] = None
    finished_at: Optional[float] = None

    model_config = DEFAULT_PYDANTIC_MODEL_CONFIG


class StockoutWatchInfo(BaseModel):
    """
    缺货监控信息
    """

    stockout_rate: Optional[float] = None
    stockout_count: Optional[int] = None
    retrieve_count: Optional[int] = None


class StorageLocationTurnoverInfo(BaseModel):
    """
    储位周转信息
    """

    # 周转数量
    turnover_quantity: float

    # 初始库存
    init_level: Optional[float] = None
    # 期末库存
    final_level: Optional[float] = None

    @property
    def turnover_rate(self) -> float:
        y = (self.init_level + self.final_level) / 2
        return self.turnover_quantity / y if y else 0
        # return self.turnover_quantity / self.capacity if self.capacity else 0


from logis.data_type import Time


class SimDurationInfo(BaseModel):
    """
    模拟运行时间信息
    """

    sim_id: str
    real: Optional[Time] = None
    sim: Optional[Time] = None
    sim_start_time: Optional[datetime] = None
    sim_end_time: Optional[datetime] = None


class RecipeItemVoAbnormalNum(RecipeItemVo):
    """配方物品异常信息"""

    abnormal_num: int  # 异常数量


class ProductionAbnormalMetric(BaseModel):
    """生产异常信息数据结构"""

    workstation_name: str  # 工作站名称
    workstation_id: Optional[str] = None  # 工作站ID
    materials: Optional[List[RecipeItemVoAbnormalNum]] = None  # 异常发生时的原材料信息


class WorkstationMetric(BaseModel):
    """工作站信息数据结构"""

    sim_id: Optional[str] = None
    workstation_id: str  # 工作站ID
    workstation_name: str  # 工作站名称
    processing_time: float  # 加工时间（秒）
    input_materials: Optional[List[RecipeItemVo]] = None  # 输入材料列表
    output_materials: Optional[List[RecipeItemVo]] = None  # 输出材料列表
    total_processing_time: float  # 使用时长
    downtime_duration: float = 0  # 停机时长（秒）


class ProductionLineMetric(BaseModel):
    """产线信息数据结构"""

    sim_id: Optional[str] = None
    production_line_name: str  # 产线名称
    production_beat: float  # 生产节拍（秒）
    total_production: int = 0  # 仿真周期生产总量
    product_unit: Optional[str] = "件"  # 产品单位
    capacity_utilization_rate: float = 0  # 产能利用率
    total_processing_time: float = 0  # 产线加工总时长（秒）
    average_utilization_rate: float = 0  # 平均利用率
    average_usage_duration: float = 0  # 平均使用时长（秒）
    workstation_num: int  # 工作站数量
    production_abnormal_list: Optional[List[ProductionAbnormalMetric]] = (
        None  # 生产异常列表
    )
    workstation_list: Optional[List[WorkstationMetric]] = None  # 工作站列表



FORMAT_DEFAULT = (
    "%(asctime)s [%(levelname).1s] %(module)s:%(funcName)s:%(lineno)d - %(message)s"
)

FORMAT_COLORED_DEFAULT = "%(asctime)s [%(log_color)s%(levelname).1s%(reset)s] %(module)s:%(funcName)s:%(lineno)d - %(blue)s%(message)s"

FORMAT_NET_DEFAULT = (
    '%(asctime)s [%(levelname).1s] %(client_addr)s - "%(request_line)s" %(status_code)s'
)

from enum import Enum


class LoggerFormat(Enum):
    """
    日志格式
    """

    DEFAULT = FORMAT_DEFAULT
    NET_DEFAULT = FORMAT_NET_DEFAULT
    COLORED_DEFAULT = FORMAT_COLORED_DEFAULT

    def formatter_name(self) -> str:
        return self.name.lower()



from abc import ABCMeta
from numbers import Number
from typing import Optional, Tuple

from logis.biz.sim.data_type import LocationType
from logis.data_type import DEFAULT_PYDANTIC_MODEL_CONFIG, NumberUnit, QuantifiedValue
from logis.task import ITask


class QuantifiedStock(QuantifiedValue):
    """
    基础货物属性
    """

    model_config = DEFAULT_PYDANTIC_MODEL_CONFIG

    sku: Optional[str] = None
    name: str
    code: Optional[str] = None

    # 默认数量为1
    quantity: Number = 1

    order_id: Optional[str] = None
    task_id: Optional[str] = None

    @property
    def unique_id(self):
        return self.sku or self.code or self.name

    @property
    def storage_key(self) -> Tuple[str, str]:
        """
        用于存储索引的key，格式为(name, unit)，
        """
        return (self.name, self.unit)


# TODO: 考虑组合实现和QuantifiedStock同源
class IStock(NumberUnit, metaclass=ABCMeta):
    # class IStock(QuantifiedStock):
    """
    库存\货物接口
    """

    order_id: Optional[str] = None
    task_id: Optional[str] = None

    name: str
    code: Optional[str] = None

    stage: Optional[str] = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def unique_id(self):
        return self.code or self.name

    @property
    def target_location(self) -> LocationType:
        raise NotImplementedError()

    @target_location.setter
    def target_location(self, location: LocationType):
        raise NotImplementedError()

    @property
    def current_location(self) -> LocationType:
        raise NotImplementedError()

    @current_location.setter
    def current_location(self, location: LocationType):
        raise NotImplementedError()

    @property
    def final_target_location(self) -> LocationType:
        """
        起初只有当前位置和目标位置，但是后来发现有些场景下目标位置只是短期目标，短期目标会经常更换
        """
        raise NotImplementedError()

    @final_target_location.setter
    def final_target_location(self, loc: LocationType):
        raise NotImplementedError()

    @property
    def storage_key(self) -> Tuple[str, str]:
        """
        用于存储索引的key，格式为(name, unit)，
        """
        return (self.name, self.unit)


class StockTask(ITask, IStock):
    pass



from setuptools import find_packages, setup

from logis import __version__

install_requires = []
with open("requirements.txt", "r") as f:
    install_requires = f.readlines()
    install_requires = [x.strip() for x in install_requires if x]
install_requires = list(filter(lambda x: not x.startswith("#"), install_requires))

setup(
    name="logis-sdk",
    version=__version__,
    description="络捷斯特 Python 模块合集",
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    author="leoking",
    author_email="present150608@sina.com",
    packages=find_packages(),
    install_requires=install_requires,
    python_requires=">=3.8",
)



from datetime import datetime
from enum import Enum
from typing import Union

import pandas as pd
from dateutil.parser import parse


class DateFormat(Enum):
    """
    日期格式枚举类
    """

    CHINESE_FULL = "%Y年%m月%d日%H点%M分%S秒"  # 2024年05月20日14点30分00秒
    CHINESE_SIMPLE = "%Y年%m月%d日"  # 2024年05月20日
    # 英文/国际标准格式
    ISO_DATETIME = "%Y-%m-%d %H:%M:%S"  # 2024-05-20 14:30:00（最常用）
    ISO_8601 = "%Y-%m-%dT%H:%M:%SZ"
    ISO_DATE = "%Y-%m-%d"  # 2024-05-20
    ISO_TIME = "%H:%M:%S"  # 14:30:00
    # 简写格式
    SHORT_DATETIME = "%y-%m-%d %H:%M"  # 24-05-20 14:30
    # 美式格式
    US_DATETIME = "%m/%d/%Y %H:%M:%S"  # 05/20/2024 14:30:00


def parse_datetime(date_str: str):
    """
    主要增加了对中文日期的解析
    """
    if not date_str:  # 处理空值
        return None

    if isinstance(date_str, (pd.Timestamp)):
        return date_str.to_pydatetime()
    
    if isinstance(date_str, (datetime)):
        return date_str


    format = [
        "%Y年%m月%d日 %H:%M:%S",  # 年月日格式
        "%Y年%m月%d日",  # 年月日格式（仅日期）
    ]
    for f in format:
        try:
            return datetime.strptime(date_str, f)
        except ValueError:
            continue
    return parse(date_str)


def format_datetime(
    dt: datetime, fmt: Union[DateFormat, str] = DateFormat.ISO_DATETIME
):
    """
    将 datetime 对象转换为中文日期格式
    """
    dt = parse_datetime(dt)
    if isinstance(fmt, DateFormat):
        fmt = fmt.value
    return dt.strftime(fmt)



from .base import *


class RetrieveStrategy(StorageStrategy):
    """
    取回策略
    """

    def find_location(
        self, storage: StorageProperties, stocks: List[QuantifiedStock], *args, **kwargs
    ) -> List[CellProperties]:
        result = []
        type_cells_map = storage.group_cell_by_stock_type()
        for stock in stocks:
            # 过滤出可以容纳此货物的所有储位
            candidates = type_cells_map.get(stock.unique_id, []) + type_cells_map.get(
                "*", []
            )
            for candidate in candidates:
                if candidate.can_retrieve(stock):
                    result.append(candidate)
        return result



from typing import Protocol, TypeVar, runtime_checkable

T = TypeVar("T")


@runtime_checkable
class IQueue(Protocol[T]):
    """
    队列接口
    """

    def __init__(self, maxsize: int = 0):
        super().__init__(maxsize)

    def put(self, item: T, *args, **kwargs):
        pass

    def put_nowait(self, item: T):
        pass

    def get(self, *args, **kwargs) -> T:
        pass

    def get_nowait(self) -> T:
        pass

    def qsize(self) -> int:
        pass

    def empty(self) -> bool:
        pass

    def full(self) -> bool:
        pass


QueueType = TypeVar("QueueType", bound=IQueue[T])



from abc import ABC
from typing import Protocol, Type, runtime_checkable

from ..data_type import ComponentConfigItem


class IComponent(ABC):
    """
    类型标识
    """

    def __init__(self, *args, **kwargs):
        pass

    def is_valid(self):
        return True


@runtime_checkable
class ComponentLoader(Protocol):
    """
    组件加载器、解析器
    """

    def load(self, item: ComponentConfigItem) -> Type[IComponent]:
        raise NotImplementedError()



from typing import Any, Callable, Dict, List, Optional, Tuple, Union

from networkx import DiGraph, Graph
from pydantic import BaseModel, Field

from logis.data_type import DEFAULT_PYDANTIC_MODEL_CONFIG, Point, Predicate

from ..base import Path, PathFindingAlgorithmType


class PathFindingConfig(BaseModel):
    obstacle_filter: Optional[Predicate[Any]] = None


class PathFindingInput(BaseModel):
    # 算法名称，分流用
    alg: Union[PathFindingAlgorithmType, str, None] = None

    # 图对象，可以是无向图或有向图
    graph: Union[Graph, DiGraph, None] = None

    # 起点
    start: Point
    # 终点
    end: Point

    # 点到点的距离矩阵
    # TODO: 从graph中获取
    distance_matrix: Dict[Point, List[Tuple[Point, float]]] = Field(
        default_factory=dict
    )
    # 排除的点或障碍点
    excluded_vertices: List[Point] = []

    # 启发式距离计算方法
    heuristic_func: Callable[[Point, Point], float] = None

    model_config = DEFAULT_PYDANTIC_MODEL_CONFIG

class PathFindingOutput(BaseModel):
    """
    寻路输出结果
    可以输出具体路径，也可以只输出下一点
    """

    # 是否成功
    success: bool

    # 具体路径的速度
    speeds: Optional[List[float]] = None

    # 可能存在多条路径
    paths: Optional[List[Path]] = None

    # 具体路径
    path: Optional[List[Point]] = None

    next_point: Optional[Point] = None

    model_config = DEFAULT_PYDANTIC_MODEL_CONFIG



import asyncio
import logging
from collections import defaultdict
from datetime import datetime
from pathlib import Path
from typing import Callable, Dict, Optional, Type
from uuid import uuid4

import simpy
from networkx import DiGraph
from sqlmodel import Session, create_engine

from logis.biz.sim import *
from logis.biz.sim.command.model import SetupArgs
from logis.biz.sim.const import SQLITE_EXT
from logis.biz.sim.data_type import SimContext
from logis.biz.sim.graph import IPathGraph

# TODO: metric是否有必要放到sdk中
from logis.biz.sim.metric import ProductionLineMetric
from logis.biz.sim.path_util import sim_data_dir
from logis.conf.app import IAppConfig
from logis.ctx import Context as BaseContext
from logis.logging import LoggerBuilder
from logis.metric import IMetricCollector
from logis.task.manager import TaskGraph


class Context(BaseContext):
    """
    仿真上下文,按照协程隔离
    """

    ID_SIM_CTX_MAP: Dict[str, SimContext] = defaultdict()
    ID_LOGGER_MAP: Dict[str, logging.Logger] = defaultdict()
    _is_simulating = False
    _simulation_lock = asyncio.Lock()

    APP_CONFIG: IAppConfig

    @classmethod
    def data_dir(cls, sim_id: Optional[str] = None) -> Path:
        """
        获取当前仿真案例所对应的数据目录
        """
        sim_id = sim_id or cls.sim_id()
        d = sim_data_dir(sim_id, data_dir=cls.APP_CONFIG.get_data_dir())
        d.mkdir(parents=True, exist_ok=True)
        return d

    @classmethod
    def data_report(
        cls, v=None, default_factory: Callable[[], "IDataReport"] = None
    ) -> "IDataReport":
        """
        获取当前仿真案例所对应的data report
        """
        key = "__data_report__"
        if v is not None:
            cls.set(key, v)
        return cls.get(key, create=True, default_factory=default_factory)

    @classmethod
    def logger_builder(cls, name: str, sim_id: Optional[str] = None) -> LoggerBuilder:
        level = cls.setup_args().level or "INFO"
        sim_id = sim_id or cls.sim_id() or ""
        return (
            LoggerBuilder()
            # 避免同名冲突
            .name(f"{sim_id}_{name}")
            .dir(cls.data_dir(sim_id=sim_id) / "logs")
            .level(level.upper())
            .backup_count(2)
        )

    @classmethod
    def logger(cls, sim_id: Optional[str] = None) -> logging.Logger:
        """
        获取当前仿真案例所对应的logger，如果当前没有仿真案例，则返回默认logger
        """
        sim_id = sim_id or cls.sim_id()
        if not sim_id:
            return cls.APP_CONFIG.get_app_logger()

        log = cls.ID_LOGGER_MAP.get(sim_id, None)
        if log:
            return log

        log = (
            cls.logger_builder(name=sim_id, sim_id=sim_id)
            .add_handler(logging.StreamHandler)
            .add_handler(
                logging.handlers.RotatingFileHandler,
                filename="info.log",
                level=logging.DEBUG,
            )
            .add_handler(
                logging.handlers.RotatingFileHandler,
                filename="warning.log",
                level=logging.WARNING,
                filters=[lambda record: record.levelno == logging.WARNING],
            )
            .add_handler(
                logging.handlers.RotatingFileHandler,
                filename="error.log",
                level=logging.ERROR,
            )
            .build()
        )
        cls.ID_LOGGER_MAP[sim_id] = log
        return log

    @classmethod
    def helper(cls) -> Dict[str, list]:
        key = "__helper__"
        return cls.get(key, default_factory=lambda: defaultdict(list), create=True)

    @classmethod
    def storage_manager(
        cls, v: Optional["IStorageManager"] = None
    ) -> "IStorageManager":
        key = "__storage_manager__"
        if v is not None:
            cls.set(key, v)
        return cls.get(key)

    @classmethod
    def add_error(cls, msg: str, sim_id: Optional[str] = None):
        cls.get_sim_context(sim_id=sim_id).errors.append(msg)

    @classmethod
    def sim_abs_start_time(cls, v: Optional[datetime] = None) -> datetime:
        """
        仿真开始的绝对时间,也就是订单表中最早任务的开始时间，也是逻辑0时刻
        """
        key = "__sim_abs_start_time__"
        return cls.get(key, default=v, create=True)

    @classmethod
    def sim_db_engine(cls, *args, **kwargs):

        def _new_db_engine():
            sim_id = cls.sim_id()
            assert sim_id, "仿真ID缺失"
            d = sim_data_dir(sim_id)
            d.mkdir(parents=True, exist_ok=True)
            db_file = d / f"{sim_id}{SQLITE_EXT}"
            return create_engine(f"sqlite:///{db_file.resolve().as_posix()}")

        key = "__db_engine_by_sim_id__"
        return cls.get(key, default_factory=_new_db_engine, create=True)

    @classmethod
    def sim_db_session(cls, sim_id: Optional[str] = None):
        return Session(cls.sim_db_engine(sim_id=sim_id))

    @classmethod
    def error_message(cls, sim_id: Optional[str] = None) -> str:
        return "\n".join((cls.get_sim_context(sim_id=sim_id).errors) or [])

    @classmethod
    def path_graph(cls, v=None) -> Optional[IPathGraph]:
        key = "__path_graph__"
        if v is not None:
            cls.set(key, v)
        return cls.get(key)

    @classmethod
    def result_generator(cls, v=None) -> Optional["IResultGenerator"]:
        key = "__result_generator__"
        if v is not None:
            cls.set(key, v)
        return cls.get(key)

    @classmethod
    def unit_manager(cls, v=None) -> Optional["IUnitManager"]:
        key = "__unit_manager__"
        if v is not None:
            cls.set(key, v)
        return cls.get(key)

    @classmethod
    def json_parser(cls, v=None) -> Optional["IJsonParser"]:
        key = "__json_parser__"
        if v is not None:
            cls.set(key, v)
        return cls.get(key)

    @classmethod
    def excel_parser(cls, v=None) -> Optional["IExcelParser"]:
        key = "__excel_parser__"
        if v is not None:
            cls.set(key, v)
        return cls.get(key)

    @classmethod
    def order_manager(cls, v=None) -> Optional["IOrderManager"]:
        key = "__order_manager__"
        if v is not None:
            cls.set(key, v)
        return cls.get(key)

    @classmethod
    def resource_manager(cls, v=None) -> Optional["IResourceManager"]:
        key = "__resource_manager__"
        if v is not None:
            cls.set(key, v)
        return cls.get(key)

    @classmethod
    def get_sim_context(
        cls, auto_create=False, sim_id: Optional[str] = None
    ) -> Optional[SimContext]:
        sim_id = sim_id or cls.sim_id()
        ctx = cls.ID_SIM_CTX_MAP.get(sim_id)
        if ctx is None and auto_create is True:
            sim_id = str(uuid4()) if sim_id is None else sim_id
            ctx = SimContext(sim_id=sim_id)
            cls.set_sim_context(ctx)
        assert sim_id, "sim_id缺失，无法找到对应上下文"
        return ctx

    @classmethod
    def set_sim_context(cls, ctx: SimContext):
        sim_id = ctx.resolve_sim_id()
        assert sim_id, "仿真上下文未初始化sim_id"
        cls.sim_id(sim_id)
        cls.ID_SIM_CTX_MAP[sim_id] = ctx

    @classmethod
    def get_env(cls, auto_create=True, sim_id: Optional[str] = None):
        ctx = cls.get_sim_context(auto_create=auto_create, sim_id=sim_id)
        return ctx.env if ctx else None

    @classmethod
    def sim_id(cls, v: Optional[str] = None, force: bool = False) -> Optional[str]:
        __key__ = "__sim_id__"
        last_id = cls.get(__key__)
        if (not force) and v is not None:
            assert (
                last_id is None or last_id == v
            ), f"上下文出现了问题，联系开发者处理: {last_id} != {v}"
            cls.set(__key__, v)
        return cls.get(__key__)

    @classmethod
    def setup_args(cls, args: Optional[SetupArgs] = None) -> SetupArgs:
        key = "__setup_args__"
        if args:
            setattr(cls, key, args)
        return getattr(cls, key, SetupArgs())

    @classmethod
    def network(cls, sim_id: Optional[str] = None) -> DiGraph:
        return cls.get_sim_context(sim_id=sim_id).network

    @classmethod
    def task_graph(cls) -> TaskGraph:
        key = "__task_graph__"
        return cls.get(key, default_factory=TaskGraph, create=True)

    @classmethod
    def rack_graph(cls) -> DiGraph:
        key = "__rack_graph__"
        return cls.get(key, default_factory=DiGraph, create=True)

    @classmethod
    def terminate_event(cls, sim_id: Optional[str] = None) -> simpy.Event:
        sim_id = sim_id or cls.sim_id()
        ctx = cls.get_sim_context(auto_create=False, sim_id=sim_id)
        assert ctx, f"sim_id={sim_id} 对应的上下文不存在"
        return ctx.terminate_event

    @classmethod
    def finish_event(cls, sim_id: Optional[str] = None) -> simpy.Event:
        sim_id = sim_id or cls.sim_id()
        ctx = cls.get_sim_context(auto_create=False, sim_id=sim_id)
        assert ctx, f"sim_id={sim_id} 对应的上下文不存在"
        return ctx.finish_event

    @classmethod
    def metric_collector(
        cls, default_factory: Optional[Type[IMetricCollector]] = None
    ) -> "IMetricCollector":
        key = "__metric_collector__"
        return cls.get(key, default_factory=default_factory, create=True)

    @classmethod
    def reset(cls, sim_id: Optional[str] = None):
        """
        重置上下文
        """
        sim_id = sim_id or cls.sim_id()
        cls.ID_SIM_CTX_MAP.pop(sim_id, None)
        cls.ID_LOGGER_MAP.pop(sim_id, None)
        # cls.logger().info("reset context:%s", cls.sim_id())
        super().reset()

    @classmethod
    def production_line_metrics(
        cls, v: Optional[Dict[str, "ProductionLineMetric"]] = None
    ) -> Dict[str, "ProductionLineMetric"]:
        """
        管理所有产线的指标数据
        """
        key = "__production_line_metrics__"
        if v is not None:
            cls.set(key, v)
        return cls.get(key, default_factory=dict, create=True)

    @classmethod
    def production_line_metric(
        cls, line_name: str, v: Optional["ProductionLineMetric"] = None
    ) -> Optional["ProductionLineMetric"]:
        """
        获取或设置特定产线的指标数据
        """
        metrics = cls.production_line_metrics()
        if v is not None:
            metrics[line_name] = v
        return metrics.get(line_name)

    @classmethod
    def is_simulating(cls, v: Optional[bool] = None) -> bool:
        """
        获取或设置仿真状态标志位

        Args:
            v: 如果提供，则设置仿真状态；否则返回当前状态

        Returns:
            当前仿真状态
        """
        if v is not None:
            cls._is_simulating = v
        return cls._is_simulating

    @classmethod
    async def try_set_simulating(cls) -> bool:
        """
        尝试设置仿真状态（原子操作，带锁保护）

        Returns:
            True 表示设置成功，False 表示已有仿真在进行
        """
        async with cls._simulation_lock:
            if cls._is_simulating:
                return False
            cls._is_simulating = True
            return True



from abc import ABC, abstractmethod
from typing import List, Optional, Union

from logis.biz.sim.const import OperationType, StorageSelectionStrategy
from logis.biz.sim.stock.model import IStock

from .device import CellClass, IRack, RackClass, RackGroupClass


class IRackSelectionStrategy(ABC):
    """
    货架选择策略，从若干货架中筛选出符合操作需求的货架
    """

    def __init__(
        self, rack_group: Union[RackGroupClass, None] = None, **kwargs
    ) -> None:
        """
        初始化货架选择策略
        Args:
            rack_group: 货架组，如果此策略仅针对特定货架组则建议传，否则忽略即可
        """
        self.rack_group = rack_group

    @abstractmethod
    def select_racks(
        self,
        operation: OperationType,
        stock: IStock,
        rack_group: Optional[RackGroupClass] = None,
        **kwargs,
    ) -> List[IRack]:
        """
        从所有可操作的货架中筛选出符合操作需求的货架
        Args:
            operation: 操作类型
            stocks: 货物列表
            rack_group: 货架组，如果不传则使用初始化时传的货架组
        Returns:
            符合操作需求的货架列表
        """
        rack_group = rack_group or self.rack_group
        assert rack_group is not None, "未传入货架组，无法选择货架"
        result: List[RackClass] = []
        for rack in rack_group.racks:
            if not rack.is_able_to(operation, stock):
                continue
            result.append(rack)
        return result


class DefaultRackSelectionStrategy(IRackSelectionStrategy):
    """
    默认货架选择策略
    """

    def select_racks(
        self,
        operation: OperationType,
        stock: IStock,
        rack_group: Optional[RackGroupClass] = None,
        strategy: Optional[StorageSelectionStrategy] = None,
        **kwargs,
    ):
        """
        默认货架选择策略
        Args:
            operation: 操作类型
            stock: 货物
            rack_group: 货架组，如果不传则使用初始化时传的货架组
            strategy: 此方式实际上不是规范的策略模式，仅仅是为了少写代码
        Returns:
            符合操作需求的货架列表
        """
        racks = super().select_racks(
            operation=operation, stock=stock, rack_group=rack_group, **kwargs
        )
        if not racks:
            return racks
        if StorageSelectionStrategy.BusyLevelAscend.matches(strategy):
            racks.sort(key=lambda x: x.current_jobs)
        elif StorageSelectionStrategy.DistanceAscend.matches(strategy):
            racks.sort(key=lambda x: x.distance_to(stock=stock, **kwargs))
        return racks


class ICellSelectionStrategy(ABC):
    """
    储位选择策略，从若干储位中筛选出符合操作需求的储位
    """

    def __init__(self, rack: Optional[RackClass] = None, **kwargs) -> None:
        """
        初始化储位选择策略
        Args:
            rack: 货架，如果此策略仅针对特定货架则建议传，否则忽略即可
        """
        self.rack = rack

    @abstractmethod
    def select_cells(
        self,
        operation: OperationType,
        stock: IStock,
        rack: Optional[RackClass] = None,
        **kwargs,
    ) -> List[CellClass]:
        """
        从所有可操作的储位中筛选出符合操作需求的储位
        Args:
            operation: 操作类型
            stocks: 货物列表
            rack: 货架，如果不传则使用初始化时传的货架
        Returns:
            符合操作需求的储位列表
        """
        pass


class DefaultCellSelectionStrategy(ICellSelectionStrategy):
    """
    默认储位选择策略
    """

    def select_cells(
        self,
        operation: OperationType,
        stock: IStock,
        rack: Optional[RackClass] = None,
        strategy: Optional[StorageSelectionStrategy] = None,
        **kwargs,
    ) -> List[CellClass]:
        rack = rack or self.rack
        assert rack is not None, "未传入货架，无法选择储位"
        cells = [
            cell for cell in rack.cells if cell.is_able_to(operation, stock, **kwargs)
        ]
        if not cells:
            return cells
        if StorageSelectionStrategy.NumberAscend.matches(strategy):
            if OperationType.Store.matches(operation):
                cells.sort(
                    key=lambda x: (
                        1 if x.is_still_universal else 0,
                        -x.target_quantity.capacity + x.target_quantity.level,
                        x.id,
                    )
                )
            else:
                cells.sort(key=lambda x: (-x.target_quantity.level, x.id))
        elif StorageSelectionStrategy.DistanceAscend.matches(strategy):
            if OperationType.Store.matches(operation):
                cells.sort(
                    key=lambda x: (
                        0 if x.is_still_universal else 1,
                        x.target_quantity.capacity - x.target_quantity.level,
                        x.id,
                    ),
                    reverse=True,
                )
            else:
                cells.sort(
                    key=lambda x: (
                        x.target_quantity.level,
                        x.id,
                    ),
                    reverse=True,
                )
        else:
            raise NotImplementedError(f"尚未支持的储位选择策略:{strategy}")
        return cells



import pandas
from ipa.decorator import deprecated


def is_prop_value_invalid(v: str):
    return pandas.isna(v) or (not v)


def is_stock_name_invalid(name: str):
    """
    判断是否是非法的货物名称
    """
    return is_prop_value_invalid(name)


@deprecated("no need to use static class")
class StockUtil:

    @staticmethod
    def is_prop_value_invalid(v: str):
        return is_prop_value_invalid(v)

    @staticmethod
    def is_stock_name_invalid(name: str):
        return is_stock_name_invalid(name)
